# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json
import os
from unittest import TestCase, mock
from urllib import parse

import pytest
import requests
import requests_unixsocket

from pylxd import client, exceptions


class TestClient(TestCase):
    """Tests for pylxd.client.Client."""

    def setUp(self):
        self.get_patcher = mock.patch("pylxd.client._APINode.get")
        self.get = self.get_patcher.start()

        self.post_patcher = mock.patch("pylxd.client._APINode.post")
        self.post = self.post_patcher.start()

        response = mock.MagicMock(status_code=200)
        response.json.return_value = {
            "metadata": {
                "auth": "trusted",
                "environment": {"storage": "zfs"},
            }
        }
        self.get.return_value = response

        post_response = mock.MagicMock(status_code=200)
        self.post.return_value = post_response

    def tearDown(self):
        self.get_patcher.stop()
        self.post_patcher.stop()

    @mock.patch("os.path.exists")
    def test_create(self, _path_exists):
        """Client creation sets default API endpoint."""
        _path_exists.return_value = False
        expected = "http+unix://%2Fvar%2Flib%2Flxd%2Funix.socket/1.0"

        an_client = client.Client()

        self.assertEqual(expected, an_client.api._api_endpoint)

    @mock.patch("os.path.exists")
    @mock.patch("os.environ")
    def test_create_with_snap_lxd(self, _environ, _path_exists):
        """Client creation sets default API endpoint."""
        _path_exists.return_value = True
        expected = "http+unix://%2Fvar%2Fsnap%2Flxd%2F" "common%2Flxd%2Funix.socket/1.0"

        an_client = client.Client()
        self.assertEqual(expected, an_client.api._api_endpoint)

    def test_create_LXD_DIR(self):
        """When LXD_DIR is set, use it in the client."""
        os.environ["LXD_DIR"] = "/lxd"
        expected = "http+unix://%2Flxd%2Funix.socket/1.0"

        an_client = client.Client()

        self.assertEqual(expected, an_client.api._api_endpoint)

    def test_create_endpoint(self):
        """Explicitly set the client endpoint."""
        endpoint = "http://lxd"
        expected = "http://lxd/1.0"

        an_client = client.Client(endpoint=endpoint)

        self.assertEqual(expected, an_client.api._api_endpoint)

    def test_create_endpoint_with_project(self):
        """Explicitly set the client endpoint."""
        response = mock.MagicMock(status_code=200)
        response.json.side_effect = [
            {
                "metadata": {
                    "auth": "untrusted",
                    "api_extensions": ["projects"],
                }
            },
        ]
        self.get.return_value = response

        endpoint = "/tmp/unix.socket"
        expected = "http+unix://%2Ftmp%2Funix.socket/1.0"

        with mock.patch("os.path.exists") as mock_exists:
            mock_exists.return_value = True
            a_client = client.Client(endpoint=endpoint, project="prj")

        self.assertEqual(a_client.api._api_endpoint, expected)
        self.assertEqual(a_client.api._project, "prj")

    def test_create_endpoint_unixsocket(self):
        """Test with unix socket endpoint."""
        endpoint = "/tmp/unix.socket"
        expected = "http+unix://%2Ftmp%2Funix.socket/1.0"

        with mock.patch("os.path.exists") as mock_exists:
            mock_exists.return_value = True
            a_client = client.Client(endpoint)

        self.assertEqual(expected, a_client.api._api_endpoint)

    def test_connection_404(self):
        """If the endpoint 404s, an exception is raised."""
        response = mock.MagicMock(status_code=404)
        self.get.return_value = response

        self.assertRaises(exceptions.ClientConnectionFailed, client.Client)

    def test_connection_failed(self):
        """If the connection fails, an exception is raised."""

        def raise_exception():
            raise requests.exceptions.ConnectionError()

        self.get.side_effect = raise_exception
        self.get.return_value = None

        self.assertRaises(exceptions.ClientConnectionFailed, client.Client)

    def test_connection_untrusted(self):
        """Client.trusted is False when certs are untrusted."""
        response = mock.MagicMock(status_code=200)
        response.json.return_value = {"metadata": {"auth": "untrusted"}}
        self.get.return_value = response

        an_client = client.Client()

        self.assertFalse(an_client.trusted)

    def test_connection_trusted(self):
        """Client.trusted is True when certs are untrusted."""
        response = mock.MagicMock(status_code=200)
        response.json.return_value = {"metadata": {"auth": "trusted"}}
        self.get.return_value = response

        an_client = client.Client()

        self.assertTrue(an_client.trusted)

    def test_server_clustered_false_no_info(self):
        """Client.server_clustered is False if the info is not available in metadata."""
        response = mock.MagicMock(status_code=200)
        response.json.return_value = {"metadata": {"environment": {}}}
        self.get.return_value = response
        a_client = client.Client()
        self.assertFalse(a_client.server_clustered)

    def test_server_clustered_false(self):
        """Client.server_clustered is False if not clustered."""
        response = mock.MagicMock(status_code=200)
        response.json.return_value = {
            "metadata": {"environment": {"server_clustered": False}}
        }
        self.get.return_value = response
        a_client = client.Client()
        self.assertFalse(a_client.server_clustered)

    def test_server_clustered_true(self):
        """Client.server_clustered is True if clustered."""
        response = mock.MagicMock(status_code=200)
        response.json.return_value = {
            "metadata": {"environment": {"server_clustered": True}}
        }
        self.get.return_value = response
        a_client = client.Client()
        self.assertTrue(a_client.server_clustered)

    def test_authenticate(self):
        """A client is authenticated."""
        response = mock.MagicMock(status_code=200)
        response.json.side_effect = [
            {"metadata": {"auth": "untrusted"}},
            {
                "metadata": {
                    "type": "client",
                    "fingerprint": "eaf55b72fc23aa516d709271df9b0116064bf8cfa009cf34c67c33ad32c2320c",
                }
            },
            {"metadata": {"auth": "trusted"}},
        ]
        self.get.return_value = response

        certs = (
            os.path.join(os.path.dirname(__file__), "lxd.crt"),
            os.path.join(os.path.dirname(__file__), "lxd.key"),
        )
        an_client = client.Client("https://lxd", cert=certs)

        an_client.authenticate("test-password")

        self.assertTrue(an_client.trusted)

    def test_authenticate_with_project(self):
        """A client is authenticated with a project."""
        response = mock.MagicMock(status_code=200)
        response.json.side_effect = [
            {
                "metadata": {
                    "auth": "untrusted",
                    "api_extensions": ["projects"],
                }
            },
            {
                "metadata": {
                    "type": "client",
                    "fingerprint": "eaf55b72fc23aa516d709271df9b0116064bf8cfa009cf34c67c33ad32c2320c",
                }
            },
            {
                "metadata": {
                    "auth": "trusted",
                    "environment": {"project": "test-proj"},
                }
            },
        ]
        self.get.return_value = response

        certs = (
            os.path.join(os.path.dirname(__file__), "lxd.crt"),
            os.path.join(os.path.dirname(__file__), "lxd.key"),
        )
        an_client = client.Client("https://lxd", cert=certs, project="test-proj")

        an_client.authenticate("test-password")

        self.assertTrue(an_client.trusted)
        self.assertEqual(an_client.host_info["environment"]["project"], "test-proj")

    def test_authenticate_project_not_supported(self):
        """A client raises an error if projects are not supported."""
        response = mock.MagicMock(status_code=200)
        response.json.return_value = {
            "metadata": {
                "auth": "untrusted",
                "api_extensions": [],
            }
        }
        self.get.return_value = response

        with pytest.raises(exceptions.ClientConnectionFailed):
            client.Client("https://lxd", project="test-proj")

    def test_authenticate_project_not_supported_but_default(self):
        """
        A client doesn't raise an error if projects are not supported and the
        default one is requested.
        """
        response = mock.MagicMock(status_code=200)
        response.json.side_effect = [
            {
                "metadata": {
                    "auth": "untrusted",
                    "api_extensions": [],
                }
            },
            {
                "metadata": {
                    "type": "client",
                    "fingerprint": "eaf55b72fc23aa516d709271df9b0116064bf8cfa009cf34c67c33ad32c2320c",
                }
            },
            {
                "metadata": {
                    "auth": "trusted",
                    "environment": {},
                }
            },
        ]
        self.get.return_value = response

        certs = (
            os.path.join(os.path.dirname(__file__), "lxd.crt"),
            os.path.join(os.path.dirname(__file__), "lxd.key"),
        )
        an_client = client.Client("https://lxd", cert=certs, project="default")

        an_client.authenticate("test-password")

        self.assertTrue(an_client.trusted)

    def test_authenticate_already_authenticated(self):
        """If the client is already authenticated, nothing happens."""
        an_client = client.Client()

        an_client.authenticate("test-password")

        self.assertTrue(an_client.trusted)

    def test_host_info(self):
        """Perform a host query."""
        an_client = client.Client()
        self.assertEqual("zfs", an_client.host_info["environment"]["storage"])

    def test_events(self):
        """The default websocket client is returned."""
        an_client = client.Client()

        ws_client = an_client.events()

        self.assertEqual("/1.0/events", ws_client.resource)

    def test_events_unix_socket(self):
        """A unix socket compatible websocket client is returned."""
        websocket_client = mock.Mock(resource=None)
        WebsocketClient = mock.Mock()
        WebsocketClient.return_value = websocket_client
        os.environ["LXD_DIR"] = "/lxd"
        an_client = client.Client()

        an_client.events(websocket_client=WebsocketClient)

        WebsocketClient.assert_called_once_with(
            "ws+unix:///lxd/unix.socket", ssl_options=None
        )

    def test_events_http(self):
        """An http compatible websocket client is returned."""
        websocket_client = mock.Mock(resource=None)
        WebsocketClient = mock.Mock()
        WebsocketClient.return_value = websocket_client
        an_client = client.Client("http://lxd.local")

        an_client.events(websocket_client=WebsocketClient)

        WebsocketClient.assert_called_once_with("ws://lxd.local", ssl_options=None)

    def test_events_https(self):
        """An https compatible websocket client is returned."""
        websocket_client = mock.Mock(resource=None)
        WebsocketClient = mock.Mock()
        WebsocketClient.return_value = websocket_client
        an_client = client.Client("https://lxd.local", cert=client.DEFAULT_CERTS)

        an_client.events(websocket_client=WebsocketClient)
        ssl_options = {
            "certfile": client.DEFAULT_CERTS.cert,
            "keyfile": client.DEFAULT_CERTS.key,
        }
        WebsocketClient.assert_called_once_with(
            "wss://lxd.local", ssl_options=ssl_options
        )

    def test_events_type_filter(self):
        """The websocket client can filter events by type."""

        an_client = client.Client()

        # from the itertools recipes documentation
        def powerset(types):
            from itertools import chain, combinations

            pwset = [combinations(types, r) for r in range(len(types) + 1)]
            return chain.from_iterable(pwset)

        event_path = "/1.0/events"

        for types in powerset(client.EventType):
            ws_client = an_client.events(event_types=set(types))

            actual_resource = parse.urlparse(ws_client.resource)
            expect_resource = parse.urlparse(event_path)

            if types and client.EventType.All not in types:
                type_csl = ",".join([t.value for t in types])
                query = parse.parse_qs(expect_resource.query)
                query.update({"type": type_csl})
                qs = parse.urlencode(query)
                expect_resource = expect_resource._replace(query=qs)

            self.assertEqual(expect_resource.path, actual_resource.path)

            if types and client.EventType.All not in types:
                qdict = parse.parse_qs(expect_resource.query)
                expect_types = set(qdict["type"][0].split(","))
                qdict = parse.parse_qs(actual_resource.query)
                actual_types = set(qdict["type"][0].split(","))

                self.assertEqual(expect_types, actual_types)
            else:
                self.assertEqual(expect_resource.query, actual_resource.query)

    def test_resources(self):
        a_client = client.Client()
        a_client.host_info["api_extensions"] = ["resources"]
        response = mock.MagicMock(status_code=200)
        response.json.return_value = {
            "metadata": {
                "cpu": {},
            }
        }
        self.get.return_value = response
        self.assertIn("cpu", a_client.resources)

    def test_resources_raises_conn_failed_exception(self):
        a_client = client.Client()
        a_client.host_info["api_extensions"] = ["resources"]
        response = mock.MagicMock(status_code=400)
        self.get.return_value = response
        with self.assertRaises(exceptions.ClientConnectionFailed):
            a_client.resources

    def test_resources_raises_api_extension_not_avail_exception(self):
        a_client = client.Client()
        a_client.host_info["api_extensions"] = []
        with self.assertRaises(exceptions.LXDAPIExtensionNotAvailable):
            a_client.resources

    def test_resources_uses_cache(self):
        a_client = client.Client()
        a_client._resource_cache = {"cpu": {}}
        # Client.__init__ calls get, reset the mock before trying
        # resources to confirm it wasn't called.
        self.get.called = False
        self.assertIn("cpu", a_client.resources)
        self.assertFalse(self.get.called)

    def test_has_api_extension(self):
        a_client = client.Client()
        a_client.host_info = {"api_extensions": ["one", "two"]}
        self.assertFalse(a_client.has_api_extension("three"))
        self.assertTrue(a_client.has_api_extension("one"))
        self.assertTrue(a_client.has_api_extension("two"))

    def test_assert_has_api_extension(self):
        a_client = client.Client()
        a_client.host_info = {"api_extensions": ["one", "two"]}
        with self.assertRaises(exceptions.LXDAPIExtensionNotAvailable) as c:
            self.assertFalse(a_client.assert_has_api_extension("three"))
        self.assertIn("three", str(c.exception))
        a_client.assert_has_api_extension("one")
        a_client.assert_has_api_extension("two")


class TestAPINode(TestCase):
    """Tests for pylxd.client._APINode."""

    def test_getattr(self):
        """API Nodes can use object notation for nesting."""
        node = client._APINode("http://test.com", mock.sentinel.session)
        new_node = node.test
        self.assertEqual("http://test.com/test", new_node._api_endpoint)

    def test_getattr_storage_pools(self):
        """API node with storage_pool should be storage-pool"""
        node = client._APINode("http://test.com", mock.sentinel.session)
        new_node = node.test.storage_pools
        self.assertEqual("http://test.com/test/storage-pools", new_node._api_endpoint)
        # other _ should stay as they were.
        new_node = node.test.some_thing
        self.assertEqual("http://test.com/test/some_thing", new_node._api_endpoint)

    def test_getitem(self):
        """API Nodes can use dict notation for nesting."""
        node = client._APINode("http://test.com", mock.sentinel.session)
        new_node = node["test"]
        self.assertEqual("http://test.com/test", new_node._api_endpoint)

    def test_getitem_leave_underscores_alone(self):
        """Bug 295 erronously changed underscores to '-' -- let's make sure
        it doens't happend again
        """
        node = client._APINode("http://test.com", mock.sentinel.session)
        new_node = node.thing["my_snapshot"]
        self.assertEqual("http://test.com/thing/my_snapshot", new_node._api_endpoint)

    def test_session(self):
        """Session is exposed on the _APINode."""
        session = requests.Session()
        node = client._APINode("http://test.com", session)
        self.assertIs(node.session, session)

    def test_session_passed_to_child(self):
        """session should be shared across path traversl"""
        parent_node = client._APINode("http+unix://test.com", mock.sentinel.session)
        child_node = parent_node.instances
        self.assertIs(parent_node.session, child_node.session)

    @mock.patch("pylxd.client.requests.Session")
    def test_get(self, Session):
        """Perform a session get."""
        response = mock.Mock(
            **{
                "status_code": 200,
                "json.return_value": {"type": "sync"},
            }
        )
        session = mock.Mock(**{"get.return_value": response})
        Session.return_value = session

        node = client._APINode("http://test.com", session)
        node.get()
        session.get.assert_called_once_with("http://test.com", timeout=None)

    @mock.patch("pylxd.client.requests.Session")
    def test_post(self, Session):
        """Perform a session post."""
        response = mock.Mock(
            **{
                "status_code": 200,
                "json.return_value": {"type": "sync"},
            }
        )
        session = mock.Mock(**{"post.return_value": response})
        Session.return_value = session
        node = client._APINode("http://test.com", session)
        node.post()
        session.post.assert_called_once_with("http://test.com", timeout=None)

    @mock.patch("pylxd.client.requests.Session")
    def test_post_200_not_sync(self, Session):
        """A status code of 200 with async request raises an exception."""
        response = mock.Mock(
            **{
                "status_code": 200,
                "json.return_value": {"type": "async"},
            }
        )
        session = mock.Mock(**{"post.return_value": response})
        Session.return_value = session
        node = client._APINode("http://test.com", session)
        self.assertRaises(exceptions.LXDAPIException, node.post)

    @mock.patch("pylxd.client.requests.Session")
    def test_post_missing_type_200(self, Session):
        """A missing response type raises an exception."""
        response = mock.Mock(
            **{
                "status_code": 200,
                "json.return_value": {},
            }
        )
        session = mock.Mock(**{"post.return_value": response})
        Session.return_value = session
        node = client._APINode("http://test.com", session)
        self.assertRaises(exceptions.LXDAPIException, node.post)

    @mock.patch("pylxd.client.requests.Session")
    def test_put(self, Session):
        """Perform a session put."""
        response = mock.Mock(
            **{
                "status_code": 200,
                "json.return_value": {"type": "sync"},
            }
        )
        session = mock.Mock(**{"put.return_value": response})
        Session.return_value = session
        node = client._APINode("http://test.com", session)
        node.put()
        session.put.assert_called_once_with("http://test.com", timeout=None)

    @mock.patch("pylxd.client.requests.Session")
    def test_patch(self, Session):
        """Perform a session patch."""
        response = mock.Mock(
            **{
                "status_code": 200,
                "json.return_value": {"type": "sync"},
            }
        )
        session = mock.Mock(**{"patch.return_value": response})
        Session.return_value = session
        node = client._APINode("http://test.com", session)
        node.patch()
        session.patch.assert_called_once_with("http://test.com", timeout=None)

    @mock.patch("pylxd.client.requests.Session")
    def test_delete(self, Session):
        """Perform a session delete."""
        response = mock.Mock(
            **{
                "status_code": 200,
                "json.return_value": {"type": "sync"},
            }
        )
        session = mock.Mock(**{"delete.return_value": response})
        Session.return_value = session
        node = client._APINode("http://test.com", session)
        node.delete()
        session.delete.assert_called_once_with("http://test.com", timeout=None)


class TestWebsocketClient(TestCase):
    """Tests for pylxd.client.WebsocketClient."""

    def test_handshake_ok(self):
        """A `message` attribute of an empty list is created."""
        ws_client = client._WebsocketClient("ws://an/fake/path")
        ws_client.handshake_ok()
        self.assertEqual([], ws_client.messages)

    def test_received_message(self):
        """A json dict is added to the messages attribute."""
        message = mock.Mock(data=json.dumps({"test": "data"}).encode("utf-8"))
        ws_client = client._WebsocketClient("ws://an/fake/path")
        ws_client.handshake_ok()
        ws_client.received_message(message)
        self.assertEqual({"test": "data"}, ws_client.messages[0])


class TestGetSessionForUrl(TestCase):
    """Tests for pylxd.client.get_session_for_url."""

    def test_session_unix_socket(self):
        """http+unix URL return a requests_unixsocket session."""
        session = client.get_session_for_url("http+unix://test.com")
        self.assertIsInstance(session, requests_unixsocket.Session)

    def test_session_http(self):
        """HTTP nodes return the default requests session."""
        session = client.get_session_for_url("http://test.com")
        self.assertIsInstance(session, requests.Session)
        self.assertNotIsInstance(session, requests_unixsocket.Session)

    def test_session_cert(self):
        """If certs are given, they're set on the Session."""
        certs = (
            os.path.join(os.path.dirname(__file__), "lxd.crt"),
            os.path.join(os.path.dirname(__file__), "lxd.key"),
        )
        session = client.get_session_for_url("http://test.com", cert=certs)
        self.assertEqual(session.cert, certs)

    def test_session_verify(self):
        """If verify is given, it's set on the Session."""
        session = client.get_session_for_url("http://test.com", verify=True)
        self.assertIs(session.verify, True)



# This code is stolen directly from lxd-images, for expediency's sake.
import atexit
import hashlib
import io
import json
import os
import shutil
import subprocess
import tarfile
import tempfile
import uuid


def find_on_path(command):
    """Is command on the executable search path?"""

    if "PATH" not in os.environ:
        return False
    path = os.environ["PATH"]
    for element in path.split(os.pathsep):
        if not element:
            continue
        filename = os.path.join(element, command)
        if os.path.isfile(filename) and os.access(filename, os.X_OK):
            return True
    return False


class Busybox:
    workdir = None

    def __init__(self):
        # Create our workdir
        self.workdir = tempfile.mkdtemp()

    def cleanup(self):
        if self.workdir:
            shutil.rmtree(self.workdir)

    def create_tarball(self, split=False):
        xz = "pxz" if find_on_path("pxz") else "xz"

        destination_tar = os.path.join(self.workdir, "busybox.tar")
        target_tarball = tarfile.open(destination_tar, "w:")

        if split:
            destination_tar_rootfs = os.path.join(self.workdir, "busybox.rootfs.tar")
            target_tarball_rootfs = tarfile.open(destination_tar_rootfs, "w:")

        metadata = {
            "architecture": os.uname()[4],
            "creation_date": int(os.stat("/bin/busybox").st_ctime),
            "properties": {
                "os": "Busybox",
                "architecture": os.uname()[4],
                "description": f"Busybox {os.uname()[4]}",
                "name": f"busybox-{os.uname()[4]}",
                # Don't overwrite actual busybox images.
                "obfuscate": str(uuid.uuid4()),
            },
        }

        # Add busybox
        with open("/bin/busybox", "rb") as fd:
            busybox_file = tarfile.TarInfo()
            busybox_file.size = os.stat("/bin/busybox").st_size
            busybox_file.mode = 0o755
            if split:
                busybox_file.name = "bin/busybox"
                target_tarball_rootfs.addfile(busybox_file, fd)
            else:
                busybox_file.name = "rootfs/bin/busybox"
                target_tarball.addfile(busybox_file, fd)

        # Add symlinks
        busybox = subprocess.Popen(
            ["/bin/busybox", "--list-full"],
            stdout=subprocess.PIPE,
            universal_newlines=True,
        )
        busybox.wait()

        for path in busybox.stdout.read().split("\n"):
            if not path.strip():
                continue
            if path.strip() == "bin/busybox":
                # Don't try and symlink to ourselves!
                continue
            symlink_file = tarfile.TarInfo()
            symlink_file.type = tarfile.SYMTYPE
            symlink_file.linkname = "/bin/busybox"
            if split:
                symlink_file.name = f"{path.strip()}"
                target_tarball_rootfs.addfile(symlink_file)
            else:
                symlink_file.name = f"rootfs/{path.strip()}"
                target_tarball.addfile(symlink_file)

        # Add directories
        for path in ("dev", "mnt", "proc", "root", "sys", "tmp"):
            directory_file = tarfile.TarInfo()
            directory_file.type = tarfile.DIRTYPE
            if split:
                directory_file.name = f"{path}"
                target_tarball_rootfs.addfile(directory_file)
            else:
                directory_file.name = f"rootfs/{path}"
                target_tarball.addfile(directory_file)

        # Add the metadata file
        metadata_yaml = (
            json.dumps(
                metadata,
                sort_keys=True,
                indent=4,
                separators=(",", ": "),
                ensure_ascii=False,
            ).encode("utf-8")
            + b"\n"
        )

        metadata_file = tarfile.TarInfo()
        metadata_file.size = len(metadata_yaml)
        metadata_file.name = "metadata.yaml"
        target_tarball.addfile(metadata_file, io.BytesIO(metadata_yaml))

        # Add an /etc/inittab; this is to work around:
        # http://lists.busybox.net/pipermail/busybox/2015-November/083618.html
        # Basically, since there are some hardcoded defaults that misbehave, we
        # just pass an empty inittab so those aren't applied, and then busybox
        # doesn't spin forever.
        inittab = tarfile.TarInfo()
        inittab.size = 1
        inittab.name = "/rootfs/etc/inittab"
        target_tarball.addfile(inittab, io.BytesIO(b"\n"))

        target_tarball.close()
        if split:
            target_tarball_rootfs.close()

        # Compress the tarball
        r = subprocess.call([xz, "-9", destination_tar])
        if r:
            raise Exception(f"Failed to compress: {destination_tar}")

        if split:
            r = subprocess.call([xz, "-9", destination_tar_rootfs])
            if r:
                raise Exception(f"Failed to compress: {destination_tar_rootfs}")
            return destination_tar + ".xz", destination_tar_rootfs + ".xz"
        else:
            return destination_tar + ".xz"


def create_busybox_image():
    busybox = Busybox()
    atexit.register(busybox.cleanup)

    path = busybox.create_tarball()

    with open(path, "rb") as fd:
        fingerprint = hashlib.sha256(fd.read()).hexdigest()

    return path, fingerprint



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from pylxd.tests import testing


class TestClusterMember(testing.PyLXDTestCase):
    """Tests for pylxd.models.ClusterMember."""

    def test_get(self):
        """A cluster member is retrieved."""
        member = self.client.cluster.members.get("an-member")

        assert member.url == "https://10.1.1.101:8443"
        assert member.architecture == "x86_64"
        assert member.roles == []

    def test_all(self):
        """All cluster members are returned."""
        members = self.client.cluster.members.all()

        self.assertIn("an-member", [m.server_name for m in members])



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os
import warnings

from pylxd import exceptions

MISSING = object()


class AttributeDict:
    """Wrap a dict making keys accessible as attributes."""

    def __init__(self, dct):
        for key, value in dct.items():
            setattr(self, key, value)

    def __iter__(self):
        return iter(self.__dict__.items())


class Attribute:
    """A metadata class for model attributes."""

    def __init__(self, validator=None, readonly=False, optional=False):
        self.validator = validator
        self.readonly = readonly
        self.optional = optional


class Manager:
    """A manager declaration.

    This class signals to the model that it will have a Manager
    attribute.
    """


class Parent:
    """A parent declaration.

    Child managers must keep a reference to their parent.
    """


class ModelType(type):
    """A Model metaclass.

    This metaclass converts the declarative Attribute style
    to attributes on the model instance itself.
    """

    def __new__(cls, name, bases, attrs):
        if "__slots__" in attrs and name != "Model":  # pragma: no cover
            raise TypeError("__slots__ should not be specified.")
        attributes = {}
        for_removal = []
        managers = []

        for base in bases:
            if hasattr(base, "__attributes__"):
                attributes.update(base.__attributes__)

        for key, val in attrs.items():
            if isinstance(val, Attribute):
                attributes[key] = val
                for_removal.append(key)
            if isinstance(val, Manager) or isinstance(val, Parent):
                managers.append(key)
                for_removal.append(key)
        for key in for_removal:
            del attrs[key]

        slots = list(attributes.keys())
        if "__slots__" in attrs:
            slots = slots + attrs["__slots__"]
        for base in bases:
            if "__slots__" in dir(base):
                slots = slots + base.__slots__
        if len(managers) > 0:
            slots = slots + managers
        attrs["__slots__"] = slots
        attrs["__attributes__"] = attributes

        return super().__new__(cls, name, bases, attrs)


# Global used to record which warnings have been issues already for unknown
# attributes.
_seen_attribute_warnings = set()


class Model(metaclass=ModelType):
    """A Base LXD object model.

    Objects fetched from the LXD API have state, which allows
    the objects to be used transactionally, with E-tag support,
    and be smart about I/O.

    The model lifecycle is this: A model's get/create methods will
    return an instance. That instance may or may not be a partial
    instance. If it is a partial instance, `sync` will be called
    and the rest of the object retrieved from the server when
    un-initialized attributes are read. When attributes are modified,
    the instance is marked as dirty. `save` will save the changes
    to the server.

    If the LXD server sends attributes that this version of pylxd is unaware of
    then a warning is printed.  By default the warning is issued ONCE and then
    supressed for every subsequent attempted setting.  The warnings can be
    completely suppressed by setting the environment variable PYLXD_WARNINGS to
    'none', or always displayed by setting the PYLXD_WARNINGS variable to
    'always'.
    """

    NotFound = exceptions.NotFound
    __slots__ = ["client", "__dirty__"]

    def __init__(self, client, **kwargs):
        self.__dirty__ = set()
        self.client = client

        for key, val in kwargs.items():
            try:
                setattr(self, key, val)
            except AttributeError:
                global _seen_attribute_warnings
                env = os.environ.get("PYLXD_WARNINGS", "").lower()
                item = f"{self.__class__.__name__}.{key}"
                if env != "always" and item in _seen_attribute_warnings:
                    continue
                _seen_attribute_warnings.add(item)
                if env == "none":
                    continue
                warnings.warn(
                    f'Attempted to set unknown attribute "{key}" '
                    f'on instance of "{self.__class__.__name__}"'
                )
        self.__dirty__.clear()

    def __getattribute__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            if name in self.__attributes__:
                self.sync()
                return super().__getattribute__(name)
            else:
                raise

    def __setattr__(self, name, value):
        if name in self.__attributes__:
            attribute = self.__attributes__[name]

            if attribute.validator is not None:
                if attribute.validator is not type(value):
                    value = attribute.validator(value)
            self.__dirty__.add(name)
        return super().__setattr__(name, value)

    def __iter__(self):
        for attr in self.__attributes__.keys():
            yield attr, getattr(self, attr)

    @property
    def dirty(self):
        return len(self.__dirty__) > 0

    def sync(self, rollback=False):
        """Sync from the server.

        When collections of objects are retrieved from the server, they
        are often partial objects. The full object must be retrieved before
        it can modified. This method is called when getattr is called on
        a non-initaliazed object.
        """
        # XXX: rockstar (25 Jun 2016) - This has the potential to step
        # on existing attributes.
        response = self.api.get()
        payload = response.json()["metadata"]
        for key, val in payload.items():
            if key not in self.__dirty__ or rollback:
                try:
                    setattr(self, key, val)
                except AttributeError:
                    # We have received an attribute from the server that we
                    # don't support in our model. Ignore this error, it
                    # doesn't hurt us.
                    pass

        # Make sure that *all* supported attributes are set, even those that
        # aren't supported by the server.
        missing_attrs = set(self.__attributes__.keys()) - set(payload.keys())
        for missing_attr in missing_attrs:
            setattr(self, missing_attr, MISSING)
        if rollback:
            self.__dirty__.clear()

    def rollback(self):
        """Reset the object from the server."""
        return self.sync(rollback=True)

    def save(self, wait=False):
        """Save data to the server.

        This method should write the new data to the server via marshalling.
        It should be a no-op when the object is not dirty, to prevent needless
        I/O.
        """
        marshalled = self.marshall()
        response = self.api.put(json=marshalled)

        if response.json()["type"] == "async" and wait:
            self.client.operations.wait_for_operation(response.json()["operation"])
        self.__dirty__.clear()

    def delete(self, wait=False):
        """Delete an object from the server."""
        response = self.api.delete()

        if response.json()["type"] == "async" and wait:
            self.client.operations.wait_for_operation(response.json()["operation"])
        self.client = None

    def marshall(self, skip_readonly=True):
        """Marshall the object in preparation for updating to the server."""
        marshalled = {}
        for key, attr in self.__attributes__.items():
            if attr.readonly and skip_readonly:
                continue
            if (not attr.optional) or (  # pragma: no branch
                attr.optional and hasattr(self, key)
            ):
                val = getattr(self, key)
                # Don't send back to the server an attribute it doesn't
                # support.
                if val is not MISSING:
                    marshalled[key] = val
        return marshalled

    def put(self, put_object, wait=False):
        """Access the PUT method directly for the object.

        This is to bypass the `save` method, and introduce a slightly saner
        approach of thinking about immuatable objects coming *from* the lXD
        server, and sending back PUTs and PATCHes.

        This method allows arbitrary puts to be attempted on the object (thus
        by passing the API attributes), but syncs the object overwriting any
        changes that may have been made to it.For a raw object return, see
        `raw_put`, which does not modify the object, and returns nothing.

        The `put_object` is the dictionary keys in the json object that is sent
        to the server for the API endpoint for the model.

        :param wait: If wait is True, then wait here until the operation
            completes.
        :type wait: bool
        :param put_object: jsonable dictionary to use as the PUT json object.
        :type put_object: dict
        :raises: :class:`pylxd.exception.LXDAPIException` on error
        """
        self.raw_put(put_object, wait)
        self.sync(rollback=True)

    def raw_put(self, put_object, wait=False):
        """Access the PUT method on the object direct, but with NO sync back.

        This accesses the PUT method for the object, but uses the `put_object`
        param to send as the JSON object.  It does NOT update the object
        afterwards, so it is effectively stale.  This is to allow a PUT when
        the object is no longer needed as it avoids another GET on the object.

        :param wait: If wait is True, then wait here until the operation
            completes.
        :type wait: bool
        :param put_object: jsonable dictionary to use as the PUT json object.
        :type put_object: dict
        :raises: :class:`pylxd.exception.LXDAPIException` on error
        """
        response = self.api.put(json=put_object)

        if response.json()["type"] == "async" and wait:
            self.client.operations.wait_for_operation(response.json()["operation"])

    def patch(self, patch_object, wait=False):
        """Access the PATCH method directly for the object.

        This is to bypass the `save` method, and introduce a slightly saner
        approach of thinking about immuatable objects coming *from* the lXD
        server, and sending back PUTs and PATCHes.

        This method allows arbitrary patches to be attempted on the object
        (thus by passing the API attributes), but syncs the object overwriting
        any changes that may have been made to it.  For a raw object return,
        see `raw_patch`, which does not modify the object, and returns nothing.

        The `patch_object` is the dictionary keys in the json object that is
        sent to the server for the API endpoint for the model.

        :param wait: If wait is True, then wait here until the operation
            completes.
        :type wait: bool
        :param patch_object: jsonable dictionary to use as the PUT json object.
        :type patch_object: dict
        :raises: :class:`pylxd.exception.LXDAPIException` on error
        """
        self.raw_patch(patch_object, wait)
        self.sync(rollback=True)

    def raw_patch(self, patch_object, wait=False):
        """Access the PATCH method on the object direct, but with NO sync back.

        This accesses the PATCH method for the object, but uses the
        `patch_object` param to send as the JSON object.  It does NOT update
        the object afterwards, so it is effectively stale.  This is to allow a
        PATCH when the object is no longer needed as it avoids another GET on
        the object.

        :param wait: If wait is True, then wait here until the operation
            completes.
        :type wait: bool
        :param patch_object: jsonable dictionary to use as the PUT json object.
        :type patch_object: dict
        :raises: :class:`pylxd.exception.LXDAPIException` on error
        """
        response = self.api.patch(json=patch_object)

        if response.json()["type"] == "async" and wait:
            self.client.operations.wait_for_operation(response.json()["operation"])



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random
import string
import unittest

import pylxd.exceptions as exceptions
from integration.testing import IntegrationTestCase


class StorageTestCase(IntegrationTestCase):
    def setUp(self):
        super().setUp()

        if not self.client.has_api_extension("storage"):
            self.skipTest("Required LXD API extension not available!")

    def create_storage_pool(self):
        # create a storage pool in the form of 'xxx1' as a dir.
        name = "".join(random.sample(string.ascii_lowercase, 3)) + "1"
        self.lxd.storage_pools.post(
            json={
                "config": {},
                "driver": "dir",
                "name": name,
            }
        )
        return name

    def delete_storage_pool(self, name):
        # delete the named storage pool
        try:
            self.lxd.storage_pools[name].delete()
        except exceptions.NotFound:
            pass


class TestStoragePools(StorageTestCase):
    """Tests for :py:class:`pylxd.models.storage_pools.StoragePools"""

    # note create and delete are tested in every method

    def test_get(self):
        name = self.create_storage_pool()
        self.addCleanup(self.delete_storage_pool, name)

        storage_pool = self.client.storage_pools.get(name)
        self.assertEqual(name, storage_pool.name)

    def test_all(self):
        name = self.create_storage_pool()
        self.addCleanup(self.delete_storage_pool, name)

        storage_pools = self.client.storage_pools.all()
        self.assertIn(name, [p.name for p in storage_pools])

    def test_save(self):
        name = self.create_storage_pool()
        self.addCleanup(self.delete_storage_pool, name)

        storage_pool = self.client.storage_pools.get(name)
        storage_pool.description = "My storage pool"
        storage_pool.save()

        p = self.client.storage_pools.get(name)
        self.assertEqual(p.description, "My storage pool")

    def test_put(self):
        name = self.create_storage_pool()
        self.addCleanup(self.delete_storage_pool, name)

        storage_pool = self.client.storage_pools.get(name)
        new_desc = "new description"
        self.assertNotEqual(storage_pool.description, new_desc)
        put_object = {
            "description": new_desc,
            "config": storage_pool.config,
        }
        storage_pool.put(put_object)
        self.assertEqual(storage_pool.description, new_desc)
        p = self.client.storage_pools.get(name)
        self.assertEqual(p.description, new_desc)

    def test_patch(self):
        name = self.create_storage_pool()
        self.addCleanup(self.delete_storage_pool, name)

        storage_pool = self.client.storage_pools.get(name)
        new_desc = "new description"
        self.assertNotEqual(storage_pool.description, new_desc)
        patch_object = {
            "description": new_desc,
        }
        storage_pool.patch(patch_object)
        self.assertEqual(storage_pool.description, new_desc)
        p = self.client.storage_pools.get(name)
        self.assertEqual(p.description, new_desc)


class TestStorageResources(StorageTestCase):
    """Tests for :py:class:`pylxd.models.storage_pools.StorageResources"""

    def test_get(self):
        name = self.create_storage_pool()
        self.addCleanup(self.delete_storage_pool, name)

        storage_pool = self.client.storage_pools.get(name)
        # just assert that it can be fetched and that a key exists
        resources = storage_pool.resources.get()
        self.assertIsInstance(resources.space, dict)
        self.assertIsInstance(resources.inodes, dict)


class TestStorageVolume(StorageTestCase):
    """Tests for :py:class:`pylxd.models.storage_pools.StorageVolume"""

    # note create and delete are tested in every method

    def create_storage_volume(self, pool):
        # note 'pool' needs to be storage_pool object or a string
        if isinstance(pool, str):
            pool = self.client.storage_pools.get(pool)
        vol_input = {
            "config": {},
            "type": "custom",
            # "pool": name,
            "name": "vol1",
        }
        volume = pool.volumes.create(vol_input)
        return volume

    def delete_storage_volume(self, pool, volume):
        # pool is either string or storage_pool object
        # volume is either a string of storage_pool object
        if isinstance(volume, str):
            if isinstance(pool, str):
                pool = self.client.storage_pools.get(pool)
            volume = pool.volumes.get("custom", volume)
        volume.delete()

    def test_create_and_get_and_delete(self):
        pool_name = self.create_storage_pool()
        self.addCleanup(self.delete_storage_pool, pool_name)

        storage_pool = self.client.storage_pools.get(pool_name)
        volume = self.create_storage_volume(storage_pool)
        vol_copy = storage_pool.volumes.get("custom", "vol1")
        self.assertEqual(vol_copy.name, volume.name)
        volume.delete()

    @unittest.skip("Can't test PUT on volumes as it doesn't make sense yet")
    def test_put(self):
        pass

    @unittest.skip("Can't test .save() on volumes yet - doesn't make sense")
    def test_save(self):
        pass

    @unittest.skip("Can't test PATCH on volumes yet - doesn't make sense")
    def test_patch(self):
        # as we're not using ZFS (and can't in these integration tests) we
        # can't really patch anything on a dir volume.
        pass



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json
import os

from pylxd import models
from pylxd.tests import testing


class TestCertificate(testing.PyLXDTestCase):
    """Tests for pylxd.models.Certificate."""

    def test_get(self):
        """A certificate is retrieved."""
        cert = self.client.certificates.get("an-certificate")

        self.assertEqual("certificate-content", cert.certificate)

    def test_all(self):
        """A certificates are returned."""
        certs = self.client.certificates.all()

        self.assertIn("an-certificate", [c.fingerprint for c in certs])

    def test_create(self):
        """A certificate is created."""
        cert_data = (
            open(os.path.join(os.path.dirname(__file__), "../../tests", "lxd.crt"))
            .read()
            .encode("utf-8")
        )
        an_certificate = self.client.certificates.create("test-password", cert_data)

        self.assertEqual(
            "eaf55b72fc23aa516d709271df9b0116064bf8cfa009cf34c67c33ad32c2320c",
            an_certificate.fingerprint,
        )

    def test_create_token(self):
        """A token is returned."""
        self.add_rule(
            {
                "text": json.dumps(
                    {
                        "type": "sync",
                        "status": "Operation created",
                        "status_code": 100,
                        "operation": "/1.0/operations/a1d77f1b-7dfb-44e0-a3a3-1dd18bd5c15a",
                        "error_code": 0,
                        "error": "",
                        "metadata": {
                            "id": "a1d77f1b-7dfb-44e0-a3a3-1dd18bd5c15a",
                            "class": "token",
                            "description": "Executing operation",
                            "created_at": "2022-06-01T19:22:21.778204449Z",
                            "updated_at": "2022-06-01T19:22:21.778204449Z",
                            "status": "Running",
                            "status_code": 103,
                            "resources": None,
                            "metadata": {
                                "addresses": [
                                    "127.0.0.1:8443",
                                    "[::1]:8443",
                                ],
                                "fingerprint": "eddaa6023f9064f94dd6fadb36c01d9af9de935efff76f4ebada5a2fda4753be",
                                "request": {
                                    "name": "foo",
                                    "type": "client",
                                    "restricted": True,
                                    "projects": ["default"],
                                    "certificate": "",
                                    "password": "",
                                    "token": True,
                                },
                                "secret": "6efac2f5de066103dc9798414e916996a8ffe3b9818608d4fe3ba175fae618ad",
                            },
                            "may_cancel": True,
                            "err": "",
                            "location": "foo",
                        },
                    },
                ),
                "method": "POST",
                "url": r"^http://pylxd.test/1.0/certificates$",
                "headers": {
                    "location": "/1.0/operations/a1d77f1b-7dfb-44e0-a3a3-1dd18bd5c15a",
                },
            },
        )

        a_token = self.client.certificates.create_token(
            name="foo", projects=["default"], restricted=True
        )

        self.assertEqual(
            "eyJjbGllbnRfbmFtZSI6ImZvbyIsImZpbmdlcnByaW50IjoiZWRkYWE2MDIzZjkwNjRmOTRkZDZmYWRiMzZjMDFkOWFmOWRlOTM1ZWZmZjc2ZjRlYmFkYTVhMmZkYTQ3NTNiZSIsImFkZHJlc3NlcyI6WyIxMjcuMC4wLjE6ODQ0MyIsIls6OjFdOjg0NDMiXSwic2VjcmV0IjoiNmVmYWMyZjVkZTA2NjEwM2RjOTc5ODQxNGU5MTY5OTZhOGZmZTNiOTgxODYwOGQ0ZmUzYmExNzVmYWU2MThhZCJ9",
            a_token,
        )

    def test_fetch(self):
        """A partial object is fully fetched."""
        an_certificate = models.Certificate(self.client, fingerprint="an-certificate")

        an_certificate.sync()

        self.assertEqual("certificate-content", an_certificate.certificate)

    def test_delete(self):
        """A certificate is deleted."""
        # XXX: rockstar (08 Jun 2016) - This just executes a code path. An
        # assertion should be added.
        an_certificate = models.Certificate(self.client, fingerprint="an-certificate")

        an_certificate.delete()



import hashlib
import json
from io import StringIO

from pylxd import exceptions, models
from pylxd.tests import testing


class TestImage(testing.PyLXDTestCase):
    """Tests for pylxd.models.Image."""

    def test_get(self):
        """An image is fetched."""
        fingerprint = hashlib.sha256(b"").hexdigest()
        a_image = models.Image.get(self.client, fingerprint)

        self.assertEqual(fingerprint, a_image.fingerprint)

    def test_get_not_found(self):
        """LXDAPIException is raised when the image isn't found."""

        def not_found(request, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$",
            }
        )

        fingerprint = hashlib.sha256(b"").hexdigest()

        self.assertRaises(
            exceptions.LXDAPIException, models.Image.get, self.client, fingerprint
        )

    def test_get_error(self):
        """LXDAPIException is raised on error."""

        def error(request, context):
            context.status_code = 500
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 500}
            )

        self.add_rule(
            {
                "text": error,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$",
            }
        )

        fingerprint = hashlib.sha256(b"").hexdigest()

        self.assertRaises(
            exceptions.LXDAPIException, models.Image.get, self.client, fingerprint
        )

    def test_get_by_alias(self):
        fingerprint = hashlib.sha256(b"").hexdigest()

        a_image = models.Image.get_by_alias(self.client, "an-alias")

        self.assertEqual(fingerprint, a_image.fingerprint)

    def test_exists(self):
        """An image is fetched."""
        fingerprint = hashlib.sha256(b"").hexdigest()

        self.assertTrue(models.Image.exists(self.client, fingerprint))

    def test_exists_by_alias(self):
        """An image is fetched."""
        self.assertTrue(models.Image.exists(self.client, "an-alias", alias=True))

    def test_not_exists(self):
        """LXDAPIException is raised when the image isn't found."""

        def not_found(request, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$",
            }
        )

        fingerprint = hashlib.sha256(b"").hexdigest()

        self.assertFalse(models.Image.exists(self.client, fingerprint))

    def test_all(self):
        """A list of all images is returned."""
        images = models.Image.all(self.client)

        self.assertEqual(1, len(images))

    def test_create(self):
        """An image is created."""
        fingerprint = hashlib.sha256(b"").hexdigest()
        a_image = models.Image.create(self.client, b"", public=True, wait=True)

        self.assertIsInstance(a_image, models.Image)
        self.assertEqual(fingerprint, a_image.fingerprint)

    def test_create_with_metadata(self):
        """An image with metadata is created."""
        fingerprint = hashlib.sha256(b"").hexdigest()
        a_image = models.Image.create(
            self.client, b"", metadata=b"", public=True, wait=True
        )

        self.assertIsInstance(a_image, models.Image)
        self.assertEqual(fingerprint, a_image.fingerprint)

    def test_create_with_metadata_streamed(self):
        """An image with metadata is created."""
        fingerprint = hashlib.sha256(b"").hexdigest()
        a_image = models.Image.create(
            self.client, StringIO(""), metadata=StringIO(""), public=True, wait=True
        )

        self.assertIsInstance(a_image, models.Image)
        self.assertEqual(fingerprint, a_image.fingerprint)

    def test_update(self):
        """An image is updated."""
        a_image = self.client.images.all()[0]
        a_image.sync()

        a_image.save()

    def test_fetch(self):
        """A partial object is fetched and populated."""
        a_image = self.client.images.all()[0]

        a_image.sync()

        self.assertEqual(1, a_image.size)

    def test_fetch_notfound(self):
        """A bogus image fetch raises LXDAPIException."""

        def not_found(request, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$",
            }
        )
        fingerprint = hashlib.sha256(b"").hexdigest()

        a_image = models.Image(self.client, fingerprint=fingerprint)

        self.assertRaises(exceptions.LXDAPIException, a_image.sync)

    def test_fetch_error(self):
        """A 500 error raises LXDAPIException."""

        def not_found(request, context):
            context.status_code = 500
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 500}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$",
            }
        )
        fingerprint = hashlib.sha256(b"").hexdigest()

        a_image = models.Image(self.client, fingerprint=fingerprint)

        self.assertRaises(exceptions.LXDAPIException, a_image.sync)

    def test_delete(self):
        """An image is deleted."""
        # XXX: rockstar (03 Jun 2016) - This just executes
        # a code path. There should be an assertion here, but
        # it's not clear how to assert that, just yet.
        a_image = self.client.images.all()[0]

        a_image.delete(wait=True)

    def test_export(self):
        """An image is exported."""
        expected = "e2943f8d0b0e7d5835f9533722a6e25f669acb8980daee378b4edb44da212f51"
        a_image = self.client.images.all()[0]

        data = a_image.export()
        data_sha = hashlib.sha256(data.read()).hexdigest()

        self.assertEqual(expected, data_sha)

    def test_export_not_found(self):
        """LXDAPIException is raised on export of bogus image."""

        def not_found(request, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855/export$",
            }
        )
        a_image = self.client.images.all()[0]

        self.assertRaises(exceptions.LXDAPIException, a_image.export)

    def test_export_error(self):
        """LXDAPIException is raised on API error."""

        def error(request, context):
            context.status_code = 500
            return json.dumps(
                {"type": "error", "error": "LOLOLOLOL", "error_code": 500}
            )

        self.add_rule(
            {
                "text": error,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855/export$",
            }
        )
        a_image = self.client.images.all()[0]

        self.assertRaises(exceptions.LXDAPIException, a_image.export)

    def test_add_alias(self):
        """Try to add an alias."""
        a_image = self.client.images.all()[0]
        a_image.add_alias("lol", "Just LOL")

        aliases = [a["name"] for a in a_image.aliases]
        self.assertTrue("lol" in aliases, "Image didn't get updated.")

    def test_add_alias_duplicate(self):
        """Adding a alias twice should raise an LXDAPIException."""

        def error(request, context):
            context.status_code = 409
            return json.dumps(
                {"type": "error", "error": "already exists", "error_code": 409}
            )

        self.add_rule(
            {
                "text": error,
                "method": "POST",
                "url": r"^http://pylxd.test/1.0/images/aliases$",
            }
        )

        a_image = self.client.images.all()[0]

        self.assertRaises(
            exceptions.LXDAPIException, a_image.add_alias, "lol", "Just LOL"
        )

    def test_remove_alias(self):
        """Try to remove an-alias."""
        a_image = self.client.images.all()[0]
        a_image.delete_alias("an-alias")

        self.assertEqual(0, len(a_image.aliases), "Alias didn't get deleted.")

    def test_remove_alias_error(self):
        """Try to remove an non existant alias."""

        def error(request, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": error,
                "method": "DELETE",
                "url": r"^http://pylxd.test/1.0/images/aliases/lol$",
            }
        )

        a_image = self.client.images.all()[0]
        self.assertRaises(exceptions.LXDAPIException, a_image.delete_alias, "lol")

    def test_remove_alias_not_in_image(self):
        """Try to remove an alias which is not in the current image."""
        a_image = self.client.images.all()[0]
        a_image.delete_alias("b-alias")

    def test_copy(self):
        """Try to copy an image to another LXD instance."""
        from pylxd.client import Client

        a_image = self.client.images.all()[0]

        client2 = Client(endpoint="http://pylxd2.test")
        copied_image = a_image.copy(client2, wait=True)
        self.assertEqual(a_image.fingerprint, copied_image.fingerprint)

    def test_copy_public(self):
        """Try to copy a public image."""
        from pylxd.client import Client

        def image_get(request, context):
            context.status_code = 200
            return json.dumps(
                {
                    "type": "sync",
                    "metadata": {
                        "aliases": [
                            {
                                "name": "an-alias",
                                "fingerprint": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                            }
                        ],
                        "architecture": "x86_64",
                        "cached": False,
                        "filename": "a_image.tar.bz2",
                        "fingerprint": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                        "public": True,
                        "properties": {},
                        "size": 1,
                        "auto_update": False,
                        "created_at": "1983-06-16T02:42:00Z",
                        "expires_at": "1983-06-16T02:42:00Z",
                        "last_used_at": "1983-06-16T02:42:00Z",
                        "uploaded_at": "1983-06-16T02:42:00Z",
                    },
                }
            )

        self.add_rule(
            {
                "text": image_get,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$",
            }
        )

        a_image = self.client.images.all()[0]
        self.assertTrue(a_image.public)

        client2 = Client(endpoint="http://pylxd2.test")
        copied_image = a_image.copy(client2, wait=True)
        self.assertEqual(a_image.fingerprint, copied_image.fingerprint)

    def test_copy_no_wait(self):
        """Try to copy and don't wait."""
        from pylxd.client import Client

        a_image = self.client.images.all()[0]

        client2 = Client(endpoint="http://pylxd2.test")
        a_image.copy(client2, public=False, auto_update=False)

    def test_create_from_simplestreams(self):
        """Try to create an image from simplestreams."""
        image = self.client.images.create_from_simplestreams(
            "https://cloud-images.ubuntu.com/releases", "trusty/amd64"
        )
        self.assertEqual(
            "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
            image.fingerprint,
        )

    def test_create_from_url(self):
        """Try to create an image from an URL."""
        image = self.client.images.create_from_url("https://dl.stgraber.org/lxd")
        self.assertEqual(
            "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
            image.fingerprint,
        )



import json


def instances_POST(request, context):
    context.status_code = 202
    return json.dumps(
        {"type": "async", "operation": "/1.0/operations/operation-abc?project=default"}
    )


def instance_POST(request, context):
    context.status_code = 202
    if not request.json().get("migration", False):
        return {
            "type": "async",
            "operation": "/1.0/operations/operation-abc?project=default",
        }
    else:
        return {
            "type": "async",
            "operation": "/1.0/operations/operation-abc?project=default",
            "metadata": {
                "metadata": {
                    "0": "abc",
                    "1": "def",
                    "control": "ghi",
                }
            },
        }


def instance_PUT(request, context):
    context.status_code = 202
    return {
        "type": "async",
        "operation": "/1.0/operations/operation-abc?project=default",
    }


def instance_DELETE(request, context):
    context.status_code = 202
    return json.dumps(
        {"type": "async", "operation": "/1.0/operations/operation-abc?project=default"}
    )


def images_POST(request, context):
    context.status_code = 202
    return json.dumps(
        {
            "type": "async",
            "operation": "/1.0/operations/images-create-operation?project=default",
        }
    )


def image_DELETE(request, context):
    context.status_code = 202
    return json.dumps(
        {"type": "async", "operation": "/1.0/operations/operation-abc?project=default"}
    )


def networks_GET(request, _):
    name = request.path.split("/")[-1]
    return json.dumps(
        {
            "type": "sync",
            "metadata": {
                "config": {
                    "ipv4.address": "10.80.100.1/24",
                    "ipv4.nat": "true",
                    "ipv6.address": "none",
                    "ipv6.nat": "false",
                },
                "name": name,
                "description": "Network description",
                "type": "bridge",
                "managed": True,
                "used_by": [],
            },
        }
    )


def networks_POST(_, context):
    context.status_code = 200
    return json.dumps({"type": "sync", "metadata": {}})


def networks_DELETE(_, context):
    context.status_code = 202
    return json.dumps(
        {"type": "sync", "operation": "/1.0/operations/operation-abc?project=default"}
    )


def profile_GET(request, context):
    name = request.path.split("/")[-1]
    return json.dumps(
        {
            "type": "sync",
            "metadata": {
                "name": name,
                "description": "An description",
                "config": {},
                "devices": {},
                "used_by": [],
            },
        }
    )


def profiles_POST(request, context):
    context.status_code = 200
    return json.dumps({"type": "sync", "metadata": {}})


def profile_DELETE(request, context):
    context.status_code = 200
    return json.dumps(
        {"type": "sync", "operation": "/1.0/operations/operation-abc?project=default"}
    )


def projects_GET(request, context):
    name = request.path.split("/")[-1]
    return json.dumps(
        {
            "type": "sync",
            "metadata": {
                "name": name,
                "description": "new project is new",
                "config": {
                    "features.images": "true",
                },
                "used_by": [],
            },
        }
    )


def projects_POST(request, context):
    context.status_code = 200
    return json.dumps({"type": "sync", "metadata": {}})


def snapshot_DELETE(request, context):
    context.status_code = 202
    return json.dumps(
        {"type": "async", "operation": "/1.0/operations/operation-abc?project=default"}
    )


RULES = [
    # General service endpoints
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": {
                    "auth": "trusted",
                    "environment": {
                        "certificate": "an-pem-cert",
                    },
                    "api_extensions": [],
                },
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0$",
    },
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": {
                    "auth": "trusted",
                    "environment": {},
                    "api_extensions": [],
                },
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd2.test/1.0$",
    },
    # Certificates
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": [
                    "http://pylxd.test/1.0/certificates/an-certificate",
                ],
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/certificates$",
    },
    {
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/certificates$",
        "headers": {
            "location": "/1.0/certificates/eaf55b72fc23aa516d709271df9b0116064bf8cfa009cf34c67c33ad32c2320c",
        },
    },
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": {
                    "certificate": "certificate-content",
                    "fingerprint": "eaf55b72fc23aa516d709271df9b0116064bf8cfa009cf34c67c33ad32c2320c",
                    "type": "client",
                },
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/certificates/eaf55b72fc23aa516d709271df9b0116064bf8cfa009cf34c67c33ad32c2320c$",
    },
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": {
                    "certificate": "certificate-content",
                    "fingerprint": "an-certificate",
                    "type": "client",
                },
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/certificates/an-certificate$",
    },
    {
        "json": {
            "type": "sync",
            "metadata": {},
        },
        "status_code": 202,
        "method": "DELETE",
        "url": r"^http://pylxd.test/1.0/certificates/an-certificate$",
    },
    # Cluster
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": {
                    "server_name": "an-member",
                    "enabled": "true",
                    "member_config": [
                        {
                            "entity": "storage-pool",
                            "name": "local",
                            "key": "source",
                            "value": "",
                            "description": '"source" property for storage pool "local"',
                        },
                        {
                            "entity": "storage-pool",
                            "name": "local",
                            "key": "volatile.initial_source",
                            "value": "",
                            "description": '"volatile.initial_source" property for'
                            ' storage pool "local"',
                        },
                    ],
                },
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/cluster$",
    },
    # Cluster Members
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": [
                    "http://pylxd.test/1.0/certificates/an-member",
                    "http://pylxd.test/1.0/certificates/nd-member",
                ],
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/cluster/members$",
    },
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": {
                    "server_name": "an-member",
                    "url": "https://10.1.1.101:8443",
                    "database": "false",
                    "status": "Online",
                    "message": "fully operational",
                    "architecture": "x86_64",
                    "description": "AMD Epyc 32c/64t",
                    "failure_domain": "rack1",
                    "roles": [],
                },
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/cluster/members/an-member$",
    },
    # cluster-certificate
    {
        "text": json.dumps({"type": "sync", "status": "Success", "status_code": 200}),
        "method": "PUT",
        "url": r"^http://pylxd.test/1.0/cluster/certificate$",
    },
    # Instances
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": [
                    "http://pylxd.test/1.0/instances/an-instance",
                ],
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/instances$",
    },
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": [
                    "http://pylxd2.test/1.0/instances/an-instance",
                ],
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd2.test/1.0/instances$",
    },
    {
        "text": instances_POST,
        "method": "POST",
        "url": r"^http://pylxd2.test/1.0/instances$",
    },
    {
        "text": instances_POST,
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/instances$",
    },
    {
        "text": instances_POST,
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/instances\?target=an-remote",
    },
    {
        "json": {
            "type": "sync",
            "metadata": {
                "name": "an-instance",
                "architecture": "x86_64",
                "config": {
                    "security.privileged": "true",
                },
                "created_at": "1983-06-16T00:00:00-00:00",
                "last_used_at": "1983-06-16T00:00:00-00:00",
                "description": "Some description",
                "devices": {"root": {"path": "/", "type": "disk"}},
                "ephemeral": True,
                "expanded_config": {
                    "security.privileged": "true",
                },
                "expanded_devices": {
                    "eth0": {
                        "name": "eth0",
                        "nictype": "bridged",
                        "parent": "lxdbr0",
                        "type": "nic",
                    },
                    "root": {"path": "/", "type": "disk"},
                },
                "profiles": ["default"],
                "stateful": False,
                "status": "Running",
                "status_code": 103,
                "unsupportedbypylxd": (
                    "This attribute is not supported by "
                    "pylxd. We want to test whether the mere presence of it "
                    "makes it crash."
                ),
            },
        },
        "method": "GET",
        "url": r"^http://pylxd2.test/1.0/instances/an-instance$",
    },
    {
        "json": {
            "type": "sync",
            "metadata": {
                "name": "an-instance",
                "architecture": "x86_64",
                "config": {
                    "security.privileged": "true",
                },
                "created_at": "1983-06-16T00:00:00-00:00",
                "last_used_at": "1983-06-16T00:00:00-00:00",
                "description": "Some description",
                "devices": {"root": {"path": "/", "type": "disk"}},
                "ephemeral": True,
                "expanded_config": {
                    "security.privileged": "true",
                },
                "expanded_devices": {
                    "eth0": {
                        "name": "eth0",
                        "nictype": "bridged",
                        "parent": "lxdbr0",
                        "type": "nic",
                    },
                    "root": {"path": "/", "type": "disk"},
                },
                "profiles": ["default"],
                "stateful": False,
                "status": "Running",
                "status_code": 103,
                "unsupportedbypylxd": (
                    "This attribute is not supported by "
                    "pylxd. We want to test whether the mere presence of it "
                    "makes it crash."
                ),
            },
        },
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/instances/an-instance$",
    },
    {
        "json": {
            "type": "sync",
            "metadata": {
                "status": "Running",
                "status_code": 103,
                "disk": {
                    "root": {
                        "usage": 10,
                    }
                },
                "memory": {
                    "usage": 15,
                    "usage_peak": 20,
                    "swap_usage": 0,
                    "swap_usage_peak": 5,
                },
                "network": {
                    "l0": {
                        "addresses": [
                            {
                                "family": "inet",
                                "address": "127.0.0.1",
                                "netmask": "8",
                                "scope": "local",
                            }
                        ],
                    }
                },
                "pid": 69,
                "processes": 100,
            },
        },
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/instances/an-instance/state$",
    },
    {
        "json": {
            "type": "sync",
            "metadata": {
                "name": "an-new-remote-instance",
                "architecture": "x86_64",
                "config": {
                    "security.privileged": "true",
                },
                "created_at": "1983-06-16T00:00:00-00:00",
                "last_used_at": "1983-06-16T00:00:00-00:00",
                "description": "Some description",
                "location": "an-remote",
                "status": "Running",
                "status_code": 103,
                "unsupportedbypylxd": (
                    "This attribute is not supported by "
                    "pylxd. We want to test whether the mere presence of it "
                    "makes it crash."
                ),
            },
        },
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/instances/an-new-remote-instance$",
    },
    {
        "status_code": 202,
        "json": {
            "type": "async",
            "operation": "/1.0/operations/operation-abc?project=default",
        },
        "method": "PUT",
        "url": r"^http://pylxd.test/1.0/instances/an-instance/state$",
    },
    {
        "json": instance_POST,
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/instances/an-instance$",
    },
    {
        "text": json.dumps(
            {
                "type": "async",
                "operation": "/1.0/operations/operation-abc?project=default",
            }
        ),
        "status_code": 202,
        "method": "PUT",
        "url": r"^http://pylxd.test/1.0/instances/an-instance$",
    },
    {
        "text": instance_DELETE,
        "method": "DELETE",
        "url": r"^http://pylxd.test/1.0/instances/an-instance$",
    },
    {
        "json": {
            "type": "async",
            "metadata": {
                "metadata": {
                    "fds": {
                        "0": "abc",
                        "1": "def",
                        "2": "ghi",
                        "control": "jkl",
                    }
                },
            },
            "operation": "/1.0/operations/operation-abc?project=default",
        },
        "status_code": 202,
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/instances/an-instance/exec$",
    },
    {
        "json": instance_PUT,
        "method": "PUT",
        "url": r"^http://pylxd.test/1.0/instances/an-instance$",
    },
    # Instance Snapshots
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": [
                    "/1.0/instances/an_instance/snapshots/an-snapshot",
                ],
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/instances/an-instance/snapshots$",
    },
    {
        "text": json.dumps(
            {
                "type": "async",
                "operation": "/1.0/operations/operation-abc?project=default",
            }
        ),
        "status_code": 202,
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/instances/an-instance/snapshots$",
    },
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": {
                    "name": "an_instance/an-snapshot",
                    "stateful": False,
                },
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/instances/an-instance/snapshots/an-snapshot$",
    },
    {
        "text": json.dumps(
            {
                "type": "async",
                "operation": "/1.0/operations/operation-abc?project=default",
            }
        ),
        "status_code": 202,
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/instances/an-instance/snapshots/an-snapshot$",
    },
    {
        "text": snapshot_DELETE,
        "method": "DELETE",
        "url": r"^http://pylxd.test/1.0/instances/an-instance/snapshots/an-snapshot$",
    },
    # Instance files
    {
        "text": "This is a getted file",
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/instances/an-instance/files\?path=%2Ftmp%2Fgetted$",
    },
    {
        "text": '{"some": "value"}',
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/instances/an-instance/files\?path=%2Ftmp%2Fjson-get$",
    },
    {
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/instances/an-instance/files\?path=%2Ftmp%2Fputted$",
    },
    {
        "method": "DELETE",
        "url": r"^http://pylxd.test/1.0/instances/an-instance/files\?path=%2Ftmp%2Fputted$",
    },
    # Images
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": [
                    "http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                ],
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/images$",
    },
    {
        "text": images_POST,
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/images$",
    },
    {
        "text": images_POST,
        "method": "POST",
        "url": r"^http://pylxd2.test/1.0/images$",
    },
    {
        "json": {
            "type": "sync",
            "status": "Success",
            "status_code": 200,
            "metadata": {
                "name": "an-alias",
                "description": "an-alias",
                "target": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
            },
        },
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/images/aliases/an-alias$",
    },
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": {
                    "aliases": [
                        {
                            "name": "an-alias",
                            "fingerprint": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                        }
                    ],
                    "architecture": "x86_64",
                    "cached": False,
                    "filename": "a_image.tar.bz2",
                    "fingerprint": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                    "public": False,
                    "properties": {},
                    "size": 1,
                    "auto_update": False,
                    "created_at": "1983-06-16T02:42:00Z",
                    "expires_at": "1983-06-16T02:42:00Z",
                    "last_used_at": "1983-06-16T02:42:00Z",
                    "uploaded_at": "1983-06-16T02:42:00Z",
                },
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$",
    },
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": {
                    "aliases": [
                        {
                            "name": "an-alias",
                            "fingerprint": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                        }
                    ],
                    "architecture": "x86_64",
                    "cached": False,
                    "filename": "a_image.tar.bz2",
                    "fingerprint": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                    "public": False,
                    "properties": {},
                    "size": 1,
                    "auto_update": False,
                    "created_at": "1983-06-16T02:42:00Z",
                    "expires_at": "1983-06-16T02:42:00Z",
                    "last_used_at": "1983-06-16T02:42:00Z",
                    "uploaded_at": "1983-06-16T02:42:00Z",
                },
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd2.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$",
    },
    {
        "text": json.dumps(
            {
                "type": "async",
                "operation": "/1.0/operations/operation-abc?project=default",
            }
        ),
        "status_code": 202,
        "method": "PUT",
        "url": r"^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$",
    },
    {
        "text": "0" * 2048,
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855/export$",
    },
    {
        "text": image_DELETE,
        "method": "DELETE",
        "url": r"^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$",
    },
    # Image Aliases
    {
        "json": {
            "type": "sync",
            "status": "Success",
            "status_code": 200,
            "metadata": {
                "name": "an-alias",
                "description": "an-alias",
                "target": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
            },
        },
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/images/aliases/an-alias$",
    },
    {
        "json": {"type": "sync", "status": "Success", "metadata": None},
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/images/aliases$",
    },
    {
        "json": {
            "type": "sync",
            "status": "Success",
            "status_code": 200,
            "metadata": None,
        },
        "method": "DELETE",
        "url": r"^http://pylxd.test/1.0/images/aliases/an-alias$",
    },
    {
        "json": {
            "type": "sync",
            "status": "Success",
            "status_code": 200,
            "metadata": None,
        },
        "method": "DELETE",
        "url": r"^http://pylxd.test/1.0/images/aliases/b-alias$",
    },
    # Images secret
    {
        "json": {
            "type": "sync",
            "status": "Success",
            "status_code": 200,
            "metadata": {"metadata": {"secret": "abcdefg"}},
        },
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855/secret$",
    },
    # Networks
    {
        "json": {
            "type": "sync",
            "metadata": [
                "http://pylxd.test/1.0/networks/lo",
                "http://pylxd.test/1.0/networks/eth0",
            ],
        },
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/networks$",
    },
    {
        "text": networks_POST,
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/networks$",
    },
    {
        "json": {
            "type": "sync",
            "metadata": {
                "name": "lo",
                "type": "loopback",
                "used_by": [],
            },
        },
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/networks/lo$",
    },
    {
        "text": networks_GET,
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/networks/eth(0|1|2)$",
    },
    {
        "text": json.dumps({"type": "sync"}),
        "method": "PUT",
        "url": r"^http://pylxd.test/1.0/networks/eth0$",
    },
    {
        "text": networks_DELETE,
        "method": "DELETE",
        "url": r"^http://pylxd.test/1.0/networks/eth0$",
    },
    # Network forwards
    {
        "json": {
            "type": "sync",
            "metadata": {
                "config": {},
                "description": "Forward description",
                "listen_address": "192.0.2.1",
                "location": "eth0",
                "ports": [
                    {
                        "description": "Port description",
                        "listen_port": "80",
                        "target_address": "192.0.2.2",
                        "target_port": "80",
                    }
                ],
            },
        },
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/networks/eth0/forwards/192.0.2.1$",
    },
    {
        "json": {
            "type": "sync",
            "metadata": {
                "config": {},
                "description": "Forward description",
                "listen_address": "192.0.2.1",
                "location": "eth0",
                "ports": [
                    {
                        "description": "Port description",
                        "listen_port": "80",
                        "target_address": "192.0.2.2",
                        "target_port": "80",
                    }
                ],
            },
        },
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/networks/eth0/forwards$",
    },
    {
        "json": {
            "type": "sync",
            "metadata": {
                "config": {},
                "description": "Updated",
                "listen_address": "192.0.2.1",
                "location": "eth0",
                "ports": [
                    {
                        "description": "Port description",
                        "listen_port": "80",
                        "target_address": "192.0.2.2",
                        "target_port": "80",
                    }
                ],
            },
        },
        "method": "PUT",
        "url": r"^http://pylxd.test/1.0/networks/eth0/forwards/192.0.2.1$",
    },
    # Storage Pools
    {
        "json": {
            "type": "sync",
            "metadata": [
                "http://pylxd.test/1.0/storage-pools/lxd",
            ],
        },
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/storage-pools$",
    },
    {
        "json": {
            "type": "sync",
            "metadata": {
                "config": {"size": "0", "source": "/var/lib/lxd/disks/lxd.img"},
                "description": "",
                "name": "lxd",
                "driver": "zfs",
                "used_by": [],
            },
        },
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/storage-pools/lxd$",
    },
    {
        "json": {"type": "sync"},
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/storage-pools$",
    },
    {
        "json": {"type": "sync"},
        "method": "DELETE",
        "url": r"^http://pylxd.test/1.0/storage-pools/lxd$",
    },
    {
        "json": {"type": "sync"},
        "method": "PUT",
        "url": r"^http://pylxd.test/1.0/storage-pools/lxd$",
    },
    {
        "json": {"type": "sync"},
        "method": "PATCH",
        "url": r"^http://pylxd.test/1.0/storage-pools/lxd$",
    },
    # Storage Resources
    {
        "json": {
            "type": "sync",
            "metadata": {
                "space": {"used": 207111192576, "total": 306027577344},
                "inodes": {"used": 3275333, "total": 18989056},
            },
        },
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/storage-pools/lxd/resources$",
    },
    # Storage Volumes
    {
        "json": {
            "type": "sync",
            "metadata": [
                "/1.0/storage-pools/default/volumes/instance/c1",
                "/1.0/storage-pools/default/volumes/instance/c2",
                "/1.0/storage-pools/default/volumes/container/c3",
                "/1.0/storage-pools/default/volumes/container/c4",
                "/1.0/storage-pools/default/volumes/virtual-machine/vm1",
                "/1.0/storage-pools/default/volumes/virtual-machine/vm2",
                "/1.0/storage-pools/default/volumes/image/i1",
                "/1.0/storage-pools/default/volumes/image/i2",
                "/1.0/storage-pools/default/volumes/custom/cu1",
            ],
        },
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/storage-pools/lxd/volumes$",
    },
    # create a sync storage volume
    {
        "json": {"type": "sync"},
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/storage-pools/lxd/volumes/custom$",
    },
    {
        "json": {
            "type": "sync",
            "status": "Success",
            "status_code": 200,
            "error_code": 0,
            "error": "",
            "metadata": {
                "type": "custom",
                "used_by": [],
                "name": "cu1",
                "config": {
                    "block.filesystem": "ext4",
                    "block.mount_options": "discard",
                    "size": "10737418240",
                },
            },
        },
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/storage-pools/lxd/volumes/custom/cu1$",
    },
    # create an async storage volume
    {
        "json": {
            "type": "async",
            "operation": "/1.0/operations/operation-abc?project=default",
        },
        "status_code": 202,
        "method": "POST",
        "url": (r"^http://pylxd.test/1.0/storage-pools/" "async-lxd/volumes/custom$"),
    },
    {
        "json": {
            "type": "sync",
            "status": "Success",
            "status_code": 200,
            "error_code": 0,
            "error": "",
            "metadata": {
                "type": "custom",
                "used_by": [],
                "name": "cu1",
                "config": {
                    "block.filesystem": "ext4",
                    "block.mount_options": "discard",
                    "size": "10737418240",
                },
            },
        },
        "method": "GET",
        "url": (
            r"^http://pylxd.test/1.0/storage-pools/" "async-lxd/volumes/custom/cu1$"
        ),
    },
    # rename a storage volume, sync
    {
        "json": {
            "type": "sync",
            "metadata": {"control": "secret1", "fs": "secret2"},
        },
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/storage-pools/lxd/volumes/custom/cu1$",
    },
    # rename a storage volume, async
    {
        "json": {
            "type": "async",
            "operation": "/1.0/operations/operation-abc?project=default",
            "metadata": {"control": "secret1", "fs": "secret2"},
        },
        "method": "POST",
        "status_code": 202,
        "url": (
            r"^http://pylxd.test/1.0/storage-pools/" "async-lxd/volumes/custom/cu1$"
        ),
    },
    {
        "json": {"type": "sync"},
        "method": "PUT",
        "url": r"^http://pylxd.test/1.0/storage-pools/lxd/volumes/custom/cu1$",
    },
    {
        "json": {"type": "sync"},
        "method": "PATCH",
        "url": r"^http://pylxd.test/1.0/storage-pools/lxd/volumes/custom/cu1$",
    },
    {
        "json": {"type": "sync"},
        "method": "DELETE",
        "url": r"^http://pylxd.test/1.0/storage-pools/lxd/volumes/custom/cu1$",
    },
    # Profiles
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": [
                    "http://pylxd.test/1.0/profiles/an-profile",
                ],
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/profiles$",
    },
    {
        "text": profiles_POST,
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/profiles$",
    },
    {
        "text": profile_GET,
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/profiles/(an-profile|an-new-profile|an-renamed-profile)$",
    },
    {
        "text": json.dumps({"type": "sync"}),
        "method": "PUT",
        "url": r"^http://pylxd.test/1.0/profiles/(an-profile|an-new-profile)$",
    },
    {
        "text": json.dumps({"type": "sync"}),
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/profiles/(an-profile|an-new-profile)$",
    },
    {
        "text": profile_DELETE,
        "method": "DELETE",
        "url": r"^http://pylxd.test/1.0/profiles/(an-profile|an-new-profile)$",
    },
    # Projects
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": [
                    "http://pylxd.test/1.0/projects/test-project",
                ],
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/projects$",
    },
    {
        "text": projects_GET,
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/projects/(test-project|new-project)$",
    },
    {
        "text": projects_POST,
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/projects$",
    },
    {
        "text": json.dumps({"type": "sync"}),
        "method": "PUT",
        "url": r"^http://pylxd.test/1.0/projects/(test-project)$",
    },
    {
        "text": json.dumps({"type": "sync"}),
        "method": "POST",
        "url": r"^http://pylxd.test/1.0/projects/(new-project)$",
    },
    {
        "text": profile_DELETE,
        "method": "DELETE",
        "url": r"^http://pylxd.test/1.0/projects/(test-project)$",
    },
    # Operations
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": {"id": "operation-abc", "metadata": {"return": 0}},
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/operations/operation-abc$",
    },
    {
        "text": json.dumps(
            {
                "type": "sync",
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/operations/operation-abc/wait$",
    },
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": {"id": "operation-abc"},
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd2.test/1.0/operations/operation-abc$",
    },
    {
        "text": json.dumps(
            {
                "type": "sync",
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd2.test/1.0/operations/operation-abc/wait$",
    },
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": {
                    "id": "images-create-operation",
                    "metadata": {
                        "fingerprint": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"
                    },
                },
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/operations/images-create-operation$",
    },
    {
        "text": json.dumps(
            {
                "type": "sync",
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd.test/1.0/operations/images-create-operation/wait$",
    },
    {
        "text": json.dumps(
            {
                "type": "sync",
                "metadata": {"id": "operation-abc"},
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd2.test/1.0/operations/images-create-operation$",
    },
    {
        "text": json.dumps(
            {
                "type": "sync",
            }
        ),
        "method": "GET",
        "url": r"^http://pylxd2.test/1.0/operations/images-create-operation/wait$",
    },
]



import json

from pylxd import exceptions, models
from pylxd.tests import testing


class TestProfile(testing.PyLXDTestCase):
    """Tests for pylxd.models.Profile."""

    def test_get(self):
        """A profile is fetched."""
        name = "an-profile"
        an_profile = models.Profile.get(self.client, name)

        self.assertEqual(name, an_profile.name)

    def test_get_not_found(self):
        """LXDAPIException is raised on unknown profiles."""

        def not_found(request, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/profiles/an-profile$",
            }
        )

        self.assertRaises(
            exceptions.LXDAPIException, models.Profile.get, self.client, "an-profile"
        )

    def test_get_error(self):
        """LXDAPIException is raised on get error."""

        def error(request, context):
            context.status_code = 500
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 500}
            )

        self.add_rule(
            {
                "text": error,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/profiles/an-profile$",
            }
        )

        self.assertRaises(
            exceptions.LXDAPIException, models.Profile.get, self.client, "an-profile"
        )

    def test_exists(self):
        name = "an-profile"

        self.assertTrue(models.Profile.exists(self.client, name))

    def test_not_exists(self):
        def not_found(request, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/profiles/an-profile$",
            }
        )

        name = "an-profile"

        self.assertFalse(models.Profile.exists(self.client, name))

    def test_all(self):
        """A list of all profiles is returned."""
        profiles = models.Profile.all(self.client)

        self.assertEqual(1, len(profiles))

    def test_create(self):
        """A new profile is created."""
        an_profile = models.Profile.create(
            self.client, name="an-new-profile", config={}, devices={}
        )

        self.assertIsInstance(an_profile, models.Profile)
        self.assertEqual("an-new-profile", an_profile.name)

    def test_rename(self):
        """A profile is renamed."""
        an_profile = models.Profile.get(self.client, "an-profile")

        an_renamed_profile = an_profile.rename("an-renamed-profile")

        self.assertEqual("an-renamed-profile", an_renamed_profile.name)

    def test_update(self):
        """A profile is updated."""
        # XXX: rockstar (03 Jun 2016) - This just executes
        # a code path. There should be an assertion here, but
        # it's not clear how to assert that, just yet.
        an_profile = models.Profile.get(self.client, "an-profile")

        an_profile.save()

        self.assertEqual({}, an_profile.config)

    def test_fetch(self):
        """A partially fetched profile is made complete."""
        an_profile = self.client.profiles.all()[0]

        an_profile.sync()

        self.assertEqual("An description", an_profile.description)

    def test_fetch_notfound(self):
        """LXDAPIException is raised on bogus profile fetches."""

        def not_found(request, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/profiles/an-profile$",
            }
        )

        an_profile = models.Profile(self.client, name="an-profile")

        self.assertRaises(exceptions.LXDAPIException, an_profile.sync)

    def test_fetch_error(self):
        """LXDAPIException is raised on fetch error."""

        def error(request, context):
            context.status_code = 500
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 500}
            )

        self.add_rule(
            {
                "text": error,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/profiles/an-profile$",
            }
        )

        an_profile = models.Profile(self.client, name="an-profile")

        self.assertRaises(exceptions.LXDAPIException, an_profile.sync)

    def test_delete(self):
        """A profile is deleted."""
        # XXX: rockstar (03 Jun 2016) - This just executes
        # a code path. There should be an assertion here, but
        # it's not clear how to assert that, just yet.
        an_profile = self.client.profiles.all()[0]

        an_profile.delete()



from pylxd.models.certificate import Certificate
from pylxd.models.cluster import Cluster, ClusterCertificate, ClusterMember
from pylxd.models.container import Container
from pylxd.models.image import Image
from pylxd.models.instance import Instance, Snapshot
from pylxd.models.network import Network, NetworkForward
from pylxd.models.operation import Operation
from pylxd.models.profile import Profile
from pylxd.models.project import Project
from pylxd.models.storage_pool import StoragePool, StorageResources, StorageVolume
from pylxd.models.virtual_machine import VirtualMachine

__all__ = [
    "Certificate",
    "Cluster",
    "ClusterCertificate",
    "ClusterMember",
    "Container",
    "Image",
    "Instance",
    "Network",
    "NetworkForward",
    "Operation",
    "Profile",
    "Project",
    "Snapshot",
    "StoragePool",
    "StorageResources",
    "StorageVolume",
    "VirtualMachine",
]



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import random
import string
import unittest
import uuid

from integration.busybox import create_busybox_image
from pylxd import exceptions
from pylxd.client import Client


class IntegrationTestCase(unittest.TestCase):
    """A base test case for pylxd integration tests."""

    def setUp(self):
        super().setUp()
        self.client = Client()
        self.lxd = self.client.api

    def generate_object_name(self):
        """Generate a random object name."""
        # Underscores are not allowed in container names.
        test = self.id().split(".")[-1].replace("_", "")
        rando = str(uuid.uuid1()).split("-")[-1]
        return f"{test}-{rando}"

    def create_container(self):
        """Create a container in lxd."""
        fingerprint, alias = self.create_image()

        name = self.generate_object_name()
        machine = {
            "name": name,
            "architecture": "2",
            "profiles": ["default"],
            "ephemeral": False,
            "config": {"limits.cpu": "2"},
            "source": {"type": "image", "alias": alias},
        }
        result = self.lxd["containers"].post(json=machine)
        operation_uuid = result.json()["operation"].split("/")[-1]
        result = self.lxd.operations[operation_uuid].wait.get()

        self.addCleanup(self.delete_container, name)
        return name

    def delete_container(self, name, enforce=False):
        """Delete a container in lxd."""
        # enforce is a hack. There's a race somewhere in the delete.
        # To ensure we don't get an infinite loop, let's count.
        count = 0
        try:
            result = self.lxd["containers"][name].delete()
        except exceptions.LXDAPIException as e:
            if e.response.status_code in (400, 404):
                return
            raise
        while enforce and result.status_code == 404 and count < 10:
            try:
                result = self.lxd["containers"][name].delete()
            except exceptions.LXDAPIException as e:
                if e.response.status_code in (400, 404):
                    return
                raise
            count += 1
        try:
            operation_uuid = result.json()["operation"].split("/")[-1]
            result = self.lxd.operations[operation_uuid].wait.get()
        except KeyError:
            pass  # 404 cases are okay.

    def create_image(self):
        """Create an image in lxd."""
        path, fingerprint = create_busybox_image()
        with open(path, "rb") as f:
            headers = {
                "X-LXD-Public": "1",
            }
            response = self.lxd.images.post(data=f.read(), headers=headers)
        operation_uuid = response.json()["operation"].split("/")[-1]
        self.lxd.operations[operation_uuid].wait.get()

        alias = self.generate_object_name()
        response = self.lxd.images.aliases.post(
            json={"description": "", "target": fingerprint, "name": alias}
        )

        self.addCleanup(self.delete_image, fingerprint)
        return fingerprint, alias

    def delete_image(self, fingerprint):
        """Delete an image in lxd."""
        try:
            self.lxd.images[fingerprint].delete()
        except exceptions.LXDAPIException as e:
            if e.response.status_code == 404:
                return
            raise

    def create_profile(self):
        """Create a profile."""
        name = self.generate_object_name()
        config = {"limits.memory": "1GB"}
        self.lxd.profiles.post(json={"name": name, "config": config})
        return name

    def delete_profile(self, name):
        """Delete a profile."""
        try:
            self.lxd.profiles[name].delete()
        except exceptions.LXDAPIException as e:
            if e.response.status_code == 404:
                return
            raise

    def create_network(self):
        # get interface name in format xxx0
        name = "".join(random.sample(string.ascii_lowercase, 3)) + "0"
        self.lxd.networks.post(
            json={
                "name": name,
                "config": {},
            }
        )
        return name

    def delete_network(self, name):
        try:
            self.lxd.networks[name].delete()
        except exceptions.NotFound:
            pass

    def assertCommon(self, response):
        """Assert common LXD responses.

        LXD responses are relatively standard. This function makes assertions
        to all those standards.
        """
        self.assertEqual(response.status_code, response.json()["status_code"])
        self.assertEqual(
            ["metadata", "operation", "status", "status_code", "type"],
            sorted(response.json().keys()),
        )



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from pylxd.models import _model as model
from pylxd.tests import testing


class Item(model.Model):
    """A fake model."""

    name = model.Attribute(readonly=True)
    age = model.Attribute(int)
    data = model.Attribute()

    @property
    def api(self):
        return self.client.api.items[self.name]


class ChildItem(Item):
    """A fake model child class."""


class TestAttributeDict:
    def test_from_dict(self):
        a = model.AttributeDict({"foo": "bar", "baz": "bza"})
        assert a.foo == "bar"
        assert a.baz == "bza"

    def test_iterable(self):
        d = {"foo": "bar", "baz": "bza"}
        a = model.AttributeDict(d)
        assert dict(a) == d


class TestModel(testing.PyLXDTestCase):
    """Tests for pylxd.model.Model."""

    def setUp(self):
        super().setUp()

        self.add_rule(
            {
                "json": {
                    "type": "sync",
                    "metadata": {
                        "name": "an-item",
                        "age": 1000,
                        "data": {"key": "val"},
                    },
                },
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/items/an-item",
            }
        )
        self.add_rule(
            {
                "json": {"type": "sync", "metadata": {}},
                "method": "put",
                "url": r"^http://pylxd.test/1.0/items/an-item",
            }
        )
        self.add_rule(
            {
                "json": {"type": "sync", "metadata": {}},
                "method": "patch",
                "url": r"^http://pylxd.test/1.0/items/an-item",
            }
        )
        self.add_rule(
            {
                "json": {"type": "sync", "metadata": {}},
                "method": "DELETE",
                "url": r"^http://pylxd.test/1.0/items/an-item",
            }
        )

    def test_init(self):
        """Initial attributes are set."""
        item = Item(self.client, name="an-item", age=15, data={"key": "val"})

        self.assertEqual(self.client, item.client)
        self.assertEqual("an-item", item.name)

    @mock.patch.dict("os.environ", {"PYLXD_WARNINGS": ""})
    @mock.patch("warnings.warn")
    def test_init_warnings_once(self, mock_warn):
        with mock.patch("pylxd.models._model._seen_attribute_warnings", new=set()):
            Item(self.client, unknown="some_value")
            mock_warn.assert_called_once_with(mock.ANY)
            Item(self.client, unknown="some_value_as_well")
            mock_warn.assert_called_once_with(mock.ANY)
            Item(self.client, unknown2="some_2nd_value")
            self.assertEqual(len(mock_warn.call_args_list), 2)

    @mock.patch.dict("os.environ", {"PYLXD_WARNINGS": "none"})
    @mock.patch("warnings.warn")
    def test_init_warnings_none(self, mock_warn):
        with mock.patch("pylxd.models._model._seen_attribute_warnings", new=set()):
            Item(self.client, unknown="some_value")
            mock_warn.assert_not_called()

    @mock.patch.dict("os.environ", {"PYLXD_WARNINGS": "always"})
    @mock.patch("warnings.warn")
    def test_init_warnings_always(self, mock_warn):
        with mock.patch("pylxd.models._model._seen_attribute_warnings", new=set()):
            Item(self.client, unknown="some_value")
            mock_warn.assert_called_once_with(mock.ANY)
            Item(self.client, unknown="some_value_as_well")
            self.assertEqual(len(mock_warn.call_args_list), 2)

    @mock.patch.dict("os.environ", {"PYLXD_WARNINGS": "none"})
    def test_init_unknown_attribute(self):
        """Unknown attributes aren't set."""
        item = Item(self.client, name="an-item", nonexistent="SRSLY")

        try:
            item.nonexistent
            self.fail("item.nonexistent did not raise AttributeError")
        except AttributeError:
            pass

    def test_init_sets_attributes_on_child_class(self):
        """Ensure that .__attributes__ is set on a child class."""
        item = Item(self.client)
        child_item = ChildItem(self.client)
        self.assertEqual(len(item.__attributes__), len(child_item.__attributes__))

    def test_unknown_attribute(self):
        """Setting unknown attributes raise an exception."""

        def set_unknown_attribute():
            item = Item(self.client, name="an-item")
            item.nonexistent = "SRSLY"

        self.assertRaises(AttributeError, set_unknown_attribute)

    def test_get_unknown_attribute(self):
        """Setting unknown attributes raise an exception."""

        def get_unknown_attribute():
            item = Item(self.client, name="an-item")
            return item.nonexistent

        self.assertRaises(AttributeError, get_unknown_attribute)

    def test_unset_attribute_sync(self):
        """Reading unavailable attributes calls sync."""
        item = Item(self.client, name="an-item")

        self.assertEqual(1000, item.age)

    def test_iter(self):
        """Test models can be iterated over."""
        item = Item(self.client, name="an-item")

        self.assertDictEqual(
            {"name": "an-item", "age": 1000, "data": {"key": "val"}}, dict(item)
        )

    def test_sync(self):
        """A sync will update attributes from the server."""
        item = Item(self.client, name="an-item")

        item.sync()

        self.assertEqual(1000, item.age)

    def test_sync_dirty(self):
        """Sync will not overwrite local attribute changes."""
        item = Item(self.client, name="an-item")

        item.age = 250
        item.sync()

        self.assertEqual(250, item.age)

    def test_rollback(self):
        """Rollback resets the object from the server."""
        item = Item(self.client, name="an-item", age=15, data={"key": "val"})

        item.age = 50
        item.rollback()

        self.assertEqual(1000, item.age)
        self.assertFalse(item.dirty)

    def test_int_attribute_validator(self):
        """Age is set properly to be an int."""
        item = Item(self.client)

        item.age = "100"

        self.assertEqual(100, item.age)

    def test_int_attribute_invalid(self):
        """TypeError is raised when data can't be converted to type."""

        def set_string():
            item = Item(self.client)
            item.age = "abc"

        self.assertRaises(ValueError, set_string)

    def test_dirty(self):
        """Changes mark the object as dirty."""
        item = Item(self.client, name="an-item", age=15, data={"key": "val"})

        item.age = 100

        self.assertTrue(item.dirty)

    def test_not_dirty(self):
        """Changes mark the object as dirty."""
        item = Item(self.client, name="an-item", age=15, data={"key": "val"})

        self.assertFalse(item.dirty)

    def test_marshall(self):
        """The object is marshalled into a dict."""
        item = Item(self.client, name="an-item", age=15, data={"key": "val"})

        result = item.marshall()

        self.assertEqual({"age": 15, "data": {"key": "val"}}, result)

    def test_delete(self):
        """The object is deleted, and client is unset."""
        item = Item(self.client, name="an-item", age=15, data={"key": "val"})

        item.delete()

        self.assertIsNone(item.client)

    def test_save(self):
        """Attributes are written to the server; object is marked clean."""
        item = Item(self.client, name="an-item", age=15, data={"key": "val"})

        item.age = 69
        item.save()

        self.assertFalse(item.dirty)

    def test_put(self):
        item = Item(self.client, name="an-item", age=15, data={"key": "val"})

        item.put({"age": 69})

        # should sync back to 1000
        self.assertEqual(item.age, 1000)

    def test_raw_put(self):
        item = Item(self.client, name="an-item", age=15, data={"key": "val"})

        item.age = 55
        item.raw_put({"age": 69})

        # should sync NOT back to 1000
        self.assertEqual(item.age, 55)

    def test_put_raw_async(self):
        self.add_rule(
            {
                "json": {
                    "type": "async",
                    "metadata": {},
                    "operation": "/1.0/items/123456789",
                },
                "status_code": 202,
                "method": "put",
                "url": r"^http://pylxd.test/1.0/items/an-item$",
            }
        )
        self.add_rule(
            {
                "json": {
                    "status": "Running",
                    "status_code": 103,
                    "type": "sync",
                    "metadata": {
                        "id": "123456789",
                        "secret": "some-long-string-of-digits",
                    },
                },
                "method": "get",
                "url": r"^http://pylxd.test/1.0/operations/123456789$",
            }
        )
        self.add_rule(
            {
                "json": {
                    "type": "sync",
                },
                "method": "get",
                "url": r"^http://pylxd.test/1.0/operations/123456789/wait$",
            }
        )
        item = Item(self.client, name="an-item", age=15, data={"key": "val"})
        item.put({"age": 69}, wait=True)

    def test_patch(self):
        item = Item(self.client, name="an-item", age=15, data={"key": "val"})

        item.patch({"age": 69})
        # should sync back to 1000
        self.assertEqual(item.age, 1000)

    def test_patch_raw_async(self):
        self.add_rule(
            {
                "json": {
                    "type": "async",
                    "metadata": {},
                    "operation": "/1.0/items/123456789",
                },
                "status_code": 202,
                "method": "patch",
                "url": r"^http://pylxd.test/1.0/items/an-item$",
            }
        )
        self.add_rule(
            {
                "json": {
                    "status": "Running",
                    "status_code": 103,
                    "type": "sync",
                    "metadata": {
                        "id": "123456789",
                        "secret": "some-long-string-of-digits",
                    },
                },
                "method": "get",
                "url": r"^http://pylxd.test/1.0/operations/123456789$",
            }
        )
        self.add_rule(
            {
                "json": {
                    "type": "sync",
                },
                "method": "get",
                "url": r"^http://pylxd.test/1.0/operations/123456789/wait$",
            }
        )
        item = Item(self.client, name="an-item", age=15, data={"key": "val"})
        item.patch({"age": 69}, wait=True)



import functools
import importlib
import inspect
from contextlib import contextmanager


class BaseManager:
    """A BaseManager class for handling collection operations."""

    @property
    def manager_for(self):  # pragma: no cover
        raise AttributeError("Manager class requires 'manager_for' attribute")

    def __init__(self, *args, **kwargs):
        manager_for = self.manager_for
        module = ".".join(manager_for.split(".")[0:-1])
        obj = manager_for.split(".")[-1]
        target_module = importlib.import_module(module)
        target = getattr(target_module, obj)

        methods = inspect.getmembers(target, predicate=inspect.ismethod)
        for name, method in methods:
            func = functools.partial(method, *args, **kwargs)
            setattr(self, name, func)
        return super().__init__()


class CertificateManager(BaseManager):
    manager_for = "pylxd.models.Certificate"


class InstanceManager(BaseManager):
    manager_for = "pylxd.models.Instance"


class ContainerManager(BaseManager):
    manager_for = "pylxd.models.Container"


class VirtualMachineManager(BaseManager):
    manager_for = "pylxd.models.VirtualMachine"


class ImageManager(BaseManager):
    manager_for = "pylxd.models.Image"


class NetworkManager(BaseManager):
    manager_for = "pylxd.models.Network"


class NetworkForwardManager(BaseManager):
    manager_for = "pylxd.models.NetworkForward"


class OperationManager(BaseManager):
    manager_for = "pylxd.models.Operation"


class ProfileManager(BaseManager):
    manager_for = "pylxd.models.Profile"


class ProjectManager(BaseManager):
    manager_for = "pylxd.models.Project"


class SnapshotManager(BaseManager):
    manager_for = "pylxd.models.Snapshot"


class StoragePoolManager(BaseManager):
    manager_for = "pylxd.models.StoragePool"


class ClusterMemberManager(BaseManager):
    manager_for = "pylxd.models.ClusterMember"


class ClusterCertificateManager(BaseManager):
    manager_for = "pylxd.models.ClusterCertificate"


class ClusterManager(BaseManager):
    manager_for = "pylxd.models.Cluster"

    def __init__(self, client, *args, **kwargs):
        super().__init__(client, *args, **kwargs)
        self._client = client
        self.members = ClusterMemberManager(client)
        self.certificate = ClusterCertificateManager(client)


@contextmanager
def web_socket_manager(manager):
    try:
        yield manager
    finally:
        manager.stop()



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import unittest

from integration.testing import IntegrationTestCase
from pylxd import exceptions


class TestContainers(IntegrationTestCase):
    """Tests for `Client.containers`."""

    def test_get(self):
        """A container is fetched by name."""
        name = self.create_container()
        self.addCleanup(self.delete_container, name)

        container = self.client.containers.get(name)

        self.assertEqual(name, container.name)

    def test_all(self):
        """A list of all containers is returned."""
        name = self.create_container()
        self.addCleanup(self.delete_container, name)

        containers = self.client.containers.all()

        self.assertIn(name, [c.name for c in containers])

    def test_create(self):
        """Creates and returns a new container."""
        _, alias = self.create_image()
        config = {
            "name": "an-container",
            "architecture": "2",
            "profiles": ["default"],
            "ephemeral": True,
            "config": {"limits.cpu": "2"},
            "source": {"type": "image", "alias": alias},
        }
        self.addCleanup(self.delete_container, config["name"])

        container = self.client.containers.create(config, wait=True)

        self.assertEqual(config["name"], container.name)


class TestContainer(IntegrationTestCase):
    """Tests for Client.Container."""

    def setUp(self):
        super().setUp()
        name = self.create_container()
        self.container = self.client.containers.get(name)

    def tearDown(self):
        super().tearDown()
        self.delete_container(self.container.name)

    def test_save(self):
        """The container is updated to a new config."""
        self.container.config["limits.cpu"] = "1"
        self.container.save(wait=True)

        self.assertEqual("1", self.container.config["limits.cpu"])
        container = self.client.containers.get(self.container.name)
        self.assertEqual("1", container.config["limits.cpu"])

    def test_rename(self):
        """The container is renamed."""
        name = "an-renamed-container"
        self.container.rename(name, wait=True)

        self.assertEqual(name, self.container.name)
        container = self.client.containers.get(name)
        self.assertEqual(name, container.name)

    def test_delete(self):
        """The container is deleted."""
        self.container.delete(wait=True)

        self.assertRaises(
            exceptions.LXDAPIException, self.client.containers.get, self.container.name
        )

    def test_start_stop(self):
        """The container is started and then stopped."""
        # NOTE: rockstar (15 Feb 2016) - I don't care for the
        # multiple assertions here, but it's a okay-ish way
        # to test what we need.
        self.container.start(wait=True)

        self.assertEqual("Running", self.container.status)
        container = self.client.containers.get(self.container.name)
        self.assertEqual("Running", container.status)

        self.container.stop(wait=True)

        self.assertEqual("Stopped", self.container.status)
        container = self.client.containers.get(self.container.name)
        self.assertEqual("Stopped", container.status)

    def test_snapshot(self):
        """A container snapshot is made, renamed, and deleted."""
        # NOTE: rockstar (15 Feb 2016) - Once again, multiple things
        # asserted in the same test.
        name = "an-snapshot"
        snapshot = self.container.snapshots.create(name, wait=True)

        self.assertEqual([name], [s.name for s in self.container.snapshots.all()])

        new_name = "an-other-snapshot"
        snapshot.rename(new_name, wait=True)

        self.assertEqual([new_name], [s.name for s in self.container.snapshots.all()])

        snapshot.delete(wait=True)

        self.assertEqual([], self.container.snapshots.all())

    def test_put_get_file(self):
        """A file is written to the container and then read."""
        filepath = "/tmp/an_file"
        data = b"abcdef"

        # raises an exception if this fails.
        self.container.files.put(filepath, data)

        contents = self.container.files.get(filepath)

        self.assertEqual(data, contents)

    def test_execute(self):
        """A command is executed on the container."""
        self.container.start(wait=True)
        self.addCleanup(self.container.stop, wait=True)

        result = self.container.execute(["echo", "test"])

        self.assertEqual(0, result.exit_code)
        self.assertEqual("test\n", result.stdout)
        self.assertEqual("", result.stderr)

    def test_execute_no_buffer(self):
        """A command is executed on the container without buffering the output."""
        self.container.start(wait=True)
        self.addCleanup(self.container.stop, wait=True)
        buffer = []

        result = self.container.execute(["echo", "test"], stdout_handler=buffer.append)

        self.assertEqual(0, result.exit_code)
        self.assertEqual("", result.stdout)
        self.assertEqual("", result.stderr)
        self.assertEqual("test\n", "".join(buffer))

    def test_execute_no_decode(self):
        """A command is executed on the container that isn't utf-8 decodable"""
        self.container.start(wait=True)
        self.addCleanup(self.container.stop, wait=True)

        result = self.container.execute(["printf", "\\xff"], decode=None)

        self.assertEqual(0, result.exit_code)
        self.assertEqual(b"\xff", result.stdout)
        self.assertEqual(b"", result.stderr)

    def test_execute_force_decode(self):
        """A command is executed and force output to ascii"""
        self.container.start(wait=True)
        self.addCleanup(self.container.stop, wait=True)

        result = self.container.execute(
            ["printf", "qu\\xe9"], decode=True, encoding="latin1"
        )

        self.assertEqual(0, result.exit_code)
        self.assertEqual("qué", result.stdout)
        self.assertEqual("", result.stderr)

    def test_execute_pipes(self):
        """A command receives data from stdin and write to stdout handler"""
        self.container.start(wait=True)
        self.addCleanup(self.container.stop, wait=True)
        test_msg = "Hello world!\n"
        stdout_msgs = []

        def stdout_handler(msg):
            stdout_msgs.append(msg)

        result = self.container.execute(
            ["cat", "-"], stdin_payload=test_msg, stdout_handler=stdout_handler
        )

        self.assertEqual(0, result.exit_code)
        # if a handler is supplied then there is no stdout in result
        self.assertEqual("", result.stdout)
        self.assertEqual("", result.stderr)
        self.assertEqual(stdout_msgs, [test_msg])

    def test_publish(self):
        """A container is published."""
        # Hack to get around mocked data
        self.container.type = "container"
        image = self.container.publish(wait=True)

        self.assertIn(
            image.fingerprint, [i.fingerprint for i in self.client.images.all()]
        )

    # COMMENT gabrik - 29/08/2018:
    # This test is commented because CRIU does NOT work
    # in LXD inside LXD

    @unittest.skip("This test is broken as it assumes particular network")
    def test_migrate_running(self):
        """A running container is migrated."""
        from pylxd.client import Client

        first_host = "https://10.0.3.111:8443/"
        second_host = "https://10.0.3.222:8443/"

        client1 = Client(endpoint=first_host, verify=False)
        client1.authenticate("password")

        client2 = Client(endpoint=second_host, verify=False)
        client2.authenticate("password")
        an_container = client1.containers.get(self.container.name)
        an_container.start(wait=True)
        an_container.sync()
        an_migrated_container = an_container.migrate(client2, wait=True)

        self.assertEqual(an_container.name, an_migrated_container.name)
        self.assertEqual(client2, an_migrated_container.client)

    @unittest.skip("This test is broken as it assumes particular network")
    def test_migrate_local_client(self):
        """Raise ValueError, cannot migrate from local connection"""
        from pylxd.client import Client

        second_host = "https://10.0.3.222:8443/"
        client2 = Client(endpoint=second_host, verify=False)
        client2.authenticate("password")

        self.assertRaises(ValueError, self.container.migrate, client2)

    @unittest.skip("This test is broken as it assumes particular network")
    def test_migrate_stopped(self):
        """A stopped container is migrated."""
        from pylxd.client import Client

        first_host = "https://10.0.3.111:8443/"
        second_host = "https://10.0.3.222:8443/"

        client1 = Client(endpoint=first_host, verify=False)
        client1.authenticate("password")

        client2 = Client(endpoint=second_host, verify=False)
        client2.authenticate("password")
        an_container = client1.containers.get(self.container.name)
        an_migrated_container = an_container.migrate(client2, wait=True)

        self.assertEqual(an_container.name, an_migrated_container.name)
        self.assertEqual(client2, an_migrated_container.client)



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json
from base64 import b64encode

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import Encoding

from pylxd.models import _model as model


class Certificate(model.Model):
    """A LXD certificate."""

    certificate = model.Attribute()
    fingerprint = model.Attribute()
    type = model.Attribute()
    name = model.Attribute()
    projects = model.Attribute()
    restricted = model.Attribute()

    @classmethod
    def get(cls, client, fingerprint):
        """Get a certificate by fingerprint."""
        response = client.api.certificates[fingerprint].get()

        return cls(client, **response.json()["metadata"])

    @classmethod
    def all(cls, client):
        """Get all certificates."""
        response = client.api.certificates.get()

        certs = []
        for cert in response.json()["metadata"]:
            fingerprint = cert.split("/")[-1]
            certs.append(cls(client, fingerprint=fingerprint))
        return certs

    @classmethod
    def create(
        cls,
        client,
        password,
        cert_data,
        cert_type="client",
        name="",
        projects=None,
        restricted=False,
    ):
        """Create a new certificate."""
        cert = x509.load_pem_x509_certificate(cert_data, default_backend())
        base64_cert = cert.public_bytes(Encoding.PEM).decode("utf-8")
        # STRIP OUT CERT META "-----BEGIN CERTIFICATE-----"
        base64_cert = "\n".join(base64_cert.split("\n")[1:-2])
        data = {
            "type": cert_type,
            "certificate": base64_cert,
            "password": password,
            "name": name,
            "restricted": restricted,
            "projects": projects,
        }
        response = client.api.certificates.post(json=data)
        location = response.headers["Location"]
        fingerprint = location.split("/")[-1]
        return cls.get(client, fingerprint)

    @classmethod
    def create_token(
        cls,
        client,
        name="",
        projects=None,
        restricted=False,
    ):
        """Create a new token."""
        data = {
            "password": "",
            "certificate": "",
            "type": "client",
            "token": True,
            "name": name,
            "restricted": restricted,
            "projects": projects,
        }
        response = client.api.certificates.post(json=data)
        metadata = response.json()["metadata"]["metadata"]

        # Assemble a token from the returned metadata
        token = {
            "client_name": name,
            "fingerprint": metadata["fingerprint"],
            "addresses": metadata["addresses"],
            "secret": metadata["secret"],
        }

        # Convert to (compact) JSON and base64 encode it
        token = json.dumps(token, separators=(",", ":"))
        return b64encode(token.encode()).decode()

    @property
    def api(self):
        return self.client.api.certificates[self.fingerprint]



import json

from pylxd import exceptions, models
from pylxd.tests import testing


class TestProject(testing.PyLXDTestCase):
    """Tests for pylxd.models.Project."""

    def test_get(self):
        """A project is fetched."""
        name = "test-project"
        a_project = models.Project.get(self.client, name)

        self.assertEqual(name, a_project.name)

    def test_get_not_found(self):
        """LXDAPIException is raised on unknown project."""

        def not_found(request, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/projects/test-project$",
            }
        )

        self.assertRaises(
            exceptions.LXDAPIException, models.Project.get, self.client, "test-project"
        )

    def test_get_error(self):
        """LXDAPIException is raised on get error."""

        def error(request, context):
            context.status_code = 500
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 500}
            )

        self.add_rule(
            {
                "text": error,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/projects/test-project$",
            }
        )

        self.assertRaises(
            exceptions.LXDAPIException, models.Project.get, self.client, "test-project"
        )

    def test_exists(self):
        name = "test-project"

        self.assertTrue(models.Project.exists(self.client, name))

    def test_not_exists(self):
        def not_found(request, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/projects/test-project$",
            }
        )

        name = "test-project"

        self.assertFalse(models.Project.exists(self.client, name))

    def test_all(self):
        """A list of all projects is returned."""
        projects = models.Project.all(self.client)

        self.assertEqual(1, len(projects))

    def test_create(self):
        """A new project is created."""

        name = "new-project"
        description = "new project is new"
        config = {
            "features.images": "true",
        }
        a_project = models.Project.create(
            self.client, name="new-project", config=config, description=description
        )

        self.assertIsInstance(a_project, models.Project)
        self.assertEqual(name, a_project.name)
        self.assertEqual(config, a_project.config)
        self.assertEqual(description, a_project.description)

    def test_update(self):
        """A project is updated."""
        a_project = models.Project.get(self.client, "test-project")

        a_project.save()

        self.assertEqual({"features.images": "true"}, a_project.config)

    def test_fetch(self):
        """A partially fetched project is made complete."""
        a_project = self.client.projects.all()[0]

        a_project.sync()

        self.assertEqual("new project is new", a_project.description)

    def test_fetch_notfound(self):
        """LXDAPIException is raised on bogus project fetches."""

        def not_found(request, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/projects/test-project$",
            }
        )

        a_project = models.Project(self.client, name="test-project")

        self.assertRaises(exceptions.LXDAPIException, a_project.sync)

    def test_fetch_error(self):
        """LXDAPIException is raised on fetch error."""

        def error(request, context):
            context.status_code = 500
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 500}
            )

        self.add_rule(
            {
                "text": error,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/projects/test-project$",
            }
        )

        a_project = models.Project(self.client, name="test-project")

        self.assertRaises(exceptions.LXDAPIException, a_project.sync)

    def test_delete(self):
        """A project is deleted."""
        a_project = self.client.projects.all()[0]

        a_project.delete()
